package org.objectweb.fdb;

import org.objectweb.fdb.api.FdbSessionInternal;
import org.objectweb.fdb.api.FdbQueryManager;
import org.objectweb.fdb.impl.meta.FdbMetaRepository;
import org.objectweb.fdb.impl.meta.ClassMetaDescription;

import java.util.Collection;

/**
 * FdbSession is the principale Fdb interface for users.
 * It allows to define class, store, read, modify or remove obects.
 */
public interface FdbSession {

    /**
     * Allows users to define structure of a persistent class.
     * @param classDescription The meta-description of the persistent class to define.
     */
    void defineClass(ClassMetaDescription classDescription);

    /**
     * Check if a persistent class is define in Fdb.
     * @param fullQualifiedClassName The full class name of the class to check.
     * @return true if there already is a class with this name define in the base.
     */
    boolean isClassDefined(String fullQualifiedClassName);

    /**
     * Get class name of a persistent instance. The user has to specify the identity of the persistent instance.
     * @param oid The identity of the persistent instance.
     * @return String : The full class name of the corresponding persistent class; null otherwise.
     */
    String getClassNameFor(Object oid);


    /**
     * Create a new identity for an instance of the given class.
     * @param objectClass The class of the expecting new persistent instance.
     * @return Object : The new identity of the expecting new persistent instance.
     */
    Object makeNewFileOid(Class objectClass);

    /**
     * Create a new identity for an instance of the given class.
     * @param objectClassName The full qualified class name of the expecting new persistent instance.
     * @return Object : The new identity of the expecting new persistent instance.
     */
    Object makeNewFileOid(String objectClassName);

    /**
     * Make a new instance of a given class persistent.
     * Also create a new identity for this instance and return it.
     * @param objectToCreate The array representation of the persistent object, according to the class meta-description.
     * @param objectClass The class of the new persistent instance.
     * @return Object : The new identity of the expecting new persistent instance.
     */
    Object createObject(Object[] objectToCreate, Class objectClass);

    /**
     * Make a new instance of a given class persistent.
     * Also create a new identity for this instance and return it.
     * @param objectToCreate The array representation of the persistent object, according to the class meta-description.
     * @param objectClassName The full qualified class name of the new persistent instance.
     * @return Object : The new identity of the expecting new persistent instance.
     */
    Object createObject(Object[] objectToCreate, String objectClassName);

    /**
     * Make a new instance of a given persistent class, with a given identity.
     * The identity must has been created before with method Object createObject(Object[] objectToCreate, Class objectClass). The class specified at this time must be the same than the instance class.
     * @param oid The identity of the new instance.
     * @param objectToCreate The array representation of the persistent object, according to the class meta-description.
     * @return Object : The identity of the expecting new persistent instance.
     */
    Object createObject(Object oid, Object[] objectToCreate);

    /**
     * Read the object of the given identity in Fdb.
     * @param oid The identity of the persistent instance.
     * @param withCache Search also in session's cache.
     * @return Object[] : The array representation of the persistent object, according to the class meta-description.
     */
    Object[] readObjectByOid(Object oid, boolean withCache);

    boolean contains(Object oid);

    /**
     * Update a persistent instance with a given identity.
     * @param oid The identity of the persistent instance.
     * @param objectToUpdate The array representation of the persistent object, according to the class meta-description.
     * @return Object : The identity of the persistent instance.
     */
    Object updateObjectByOid(Object oid, Object[] objectToUpdate);

    /**
     * Remove a persistent object of the given identity from Fdb.
     * @param oid The identity of the persistent instance.
     */
    void deleteObjectByOid(Object oid);


    /**
     * Get identities of all persistent instance of a given class.
     * @param fullCLassName The full name of the class.
     * @return Collection : A collection of all the oid of the persistent instances of the class.
     */
    Collection getExtent(String fullCLassName);

    /**
     * Get identities of all persistent instance of a given class.
     * @param theClass The class.
     * @return Collection : A collection of all the oid of the persistent instances of the class.
     */
    Collection getExtent(Class theClass);


    /**
     * Tell if the fdb base has been created during the initialisation of the FdbSession instance.
     * @return boolean : true if the base has been created during the initialisation of FdbSession instance; false otherwise.
     */
    boolean isNewBase();

    /**
     * Try to optimize space in base files.
     */
    void restructureBase();

    /**
     * The getQueryManager method returns the FdbQueryManager instance associated with the FdbSession instance.
     * @return FdbQueryManager : The FdbQueryManager instance associated with the FdbSession instance.
     */
    FdbQueryManager getQueryManager();

    /**
     * The currentTransaction method returns the FdbTransaction instance associated with the FdbSession instance.
     * @return FdbTransaction : The FdbTransaction instance associated with the FdbSession instance.
     */
    FdbTransaction currentTransaction();

    /**
     * Get a MetaRepository instance to work on Fdb schema of persistent classes.
     * @return MetaRepository : the meta repository associated with FdbSession instance.
     */
    FdbMetaRepository getMetaRepository();

    /**
     * Get an internal view of FdbSession instance.
     * Users don't have to use this interface for commun functionalities.
     * @return FdbSessionInternal : A internal view of the FdbSession instance.
     */
    FdbSessionInternal getInternalSession();
}
