package org.objectweb.fdb;

import org.objectweb.fdb.api.*;
import org.objectweb.fdb.exception.FdbDatabaseException;
import org.objectweb.fdb.impl.meta.FdbMetaRepository;
import org.objectweb.fdb.impl.meta.ClassMetaDescription;
import org.objectweb.fdb.impl.*;

import java.io.PrintWriter;
import java.util.Collection;
import java.util.Iterator;
import java.util.StringTokenizer;

/**
 * Utility class to look into a Fdb base, specified by his properties file.
 * Call main method with no argument to print usage.
 */
public class FdbViewer {

    public static void main(String[] args) {
        try {
            PrintWriter out = new PrintWriter(System.out);
            FdbViewer viewer = new FdbViewer(args, out);
            viewer.readBase();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    private static void printUsage(PrintWriter writer, String message) {
        if ((message != null) && (!message.equals(""))) {
            writer.println(message);
        }
        writer.println("java org.objectweb.filestore.database.admin.FileViewer <option> : print instances in specified base");
        writer.println("\t" + FILE_OPTION + " <path> : specify the properties file of the base to inspect (mandatory)");
        writer.println("\t" + NO_INSTANCE_OPTION + " : doesn't show instance");
        writer.println("\t" + INSTANCE_OPTION + " <coma-separated floid list> : show only the floid-related instances");
        writer.println("\t" + META_OPTION + " : show the class meta-model");
        writer.println("\t" + STATE_OPTION + " : show the initial state of the base");
    }


    public FdbViewer(String[] args, PrintWriter out) {
        try {
            parseArgs(args);
        } catch (Exception e) {
            printUsage(out, e.getMessage());
            out.flush();
            System.exit(0);
        }

        this.writer = out;
        props = new FdbProperties(propFilePath);
        FdbSession session = InternalFdbFactory.getNewSession(propFilePath);
        fileAccess = InternalFdbFactory.getLogicalAccess(props);
        oidTable = InternalFdbFactory.getOidTable(props);
        repository = InternalFdbFactory.getMetaRepository((FdbSessionInternal) session, props);
    }

    public void readBase() {
        if (withState) {
            printState(0);
        }
        Iterator classNameIterator = getAllClassNames().iterator();
        while (classNameIterator.hasNext()) {
            String className = (String) classNameIterator.next();
            Collection oids = getExtent(className);
            if (withMeta) {
                printClassMeta(1, className, oids);
            }
            if ((!withSelectedInstance) && (!noInstance)) {
                printClassInstances(1, className, oids);
            }
        }
        if (withSelectedInstance && (!noInstance)) {
            StringTokenizer tokenizer = new StringTokenizer(instanceList, ",");
            while (tokenizer.hasMoreTokens()) {
                long l = Long.parseLong(tokenizer.nextToken());
                FdbLoid floid = new FdbLoid(l);
                FdbPoid fpoid = oidTable.getFpoid(floid);
                if (fpoid != null) {
                    ClassMetaDescription meta =
                            repository.getPersistentClassMetaDescription(fpoid.getClid());
                    printInstance(1, meta, floid);
                } else {
                    printInstance(1, null, floid);
                }
            }
        }
        writer.flush();
    }

    private Collection getAllClassNames() {
        return repository.getAllClassNames();
    }

    private Collection getExtent(String fullCLassName) {
        Object clid = repository.getClidFromClassName(fullCLassName);
        return oidTable.getAllFloidFromClid(clid);
    }

    private String getTab(int nbrTab) {
        String tab = "";
        for (int i = 0; i < nbrTab; i++) {
            tab = tab + "\t";
        }
        return tab;
    }

    private static void parseArgs(String[] args) {
        if (args.length == 0) {
            throw new FdbDatabaseException("FileViewer : no arguments");
        }

        boolean hasFileOption = false;

        for (int i = 0; i < args.length; i++) {
            String o = args[i];

            if (o.equalsIgnoreCase(FILE_OPTION)) {
                propFilePath = args[i + 1];
                hasFileOption = true;
                i++;
            } else if (o.equalsIgnoreCase(INSTANCE_OPTION)) {
                instanceList = args[i + 1];
                i++;
                withSelectedInstance = true;
            } else if (o.equalsIgnoreCase(NO_INSTANCE_OPTION)) {
                noInstance = true;
            } else if (o.equalsIgnoreCase(META_OPTION)) {
                withMeta = true;
            } else if (o.equalsIgnoreCase(STATE_OPTION)) {
                withState = true;
            } else if (o.equalsIgnoreCase(HELP_OPTION)) {
                throw new FdbDatabaseException("FileViewer : print help option");
            } else {
                throw new FdbDatabaseException("FileViewer : invalid argument format : " + o);
            }
        }

        if (!hasFileOption) {
            throw new FdbDatabaseException("FileViewer : no -file option");
        }
    }

    private void printClassInstances(int nbrTab, String className, Collection oids) {
        Iterator oidsIt = oids.iterator();
        Object clid = repository.getClidFromClassName(className);
        ClassMetaDescription meta = repository.getPersistentClassMetaDescription(clid);
        while (oidsIt.hasNext()) {
            FdbLoid floid = (FdbLoid) oidsIt.next();
            printInstance(nbrTab, meta, floid);
        }
        writer.println();
    }

    private void printClassMeta(int nbrTab, String className, Collection oids) {
        String tab = getTab(nbrTab);
        String tabPlus1 = getTab(nbrTab + 1);
        Object clid = repository.getClidFromClassName(className);
        ClassMetaDescription meta = repository.getPersistentClassMetaDescription(clid);
        String nbrInstanceMessage;
        int size = oids.size();
        if (size == 0) {
            nbrInstanceMessage = "no instances";
        } else if (size == 1) {
            nbrInstanceMessage = "1 instance";
        } else {
            nbrInstanceMessage = String.valueOf(size) + " instances";
        }
        writer.println(className + " : " + nbrInstanceMessage);
        writer.println(tab + "Meta-description of class " + className);
        writer.println(tabPlus1 + "Instance number : " + nbrInstanceMessage);
        String[] fieldNames = meta.getFieldNames();
        for (int i = 0; i < fieldNames.length; i++) {
            writer.println(tabPlus1 + "field " + i + " name : " + fieldNames[i]);
        }
        writer.println();
    }

    private void printInstance(int nbrTab, ClassMetaDescription meta, FdbLoid floid) {
        String tab = getTab(nbrTab);
        String tabPlus1 = getTab(nbrTab + 1);
        if (meta == null) {
            writer.println(tab + "Instance number : " + floid);
            writer.println(tabPlus1 + "object null in base");
            writer.println();
            return;
        }
        writer.println(tab + "Instance of " + meta.getClassName() + " number : " + floid);
        String[] fields = meta.getFieldNames();
        Object[] o = readObjectByFloid(floid);
        if (o != null) {
            for (int i = 0; i < o.length; i++) {
                String objectInString;
                if (o[i] instanceof String) {
                    objectInString = "\"" + o[i] + "\"";
                } else {
                    objectInString = String.valueOf(o[i]);
                }
                writer.println(tabPlus1 + fields[i] + " : " + objectInString);
            }
        } else {
            writer.println(tabPlus1 + "object null in base");
        }
        writer.println();
    }

    private void printState(int nbrTab) {
        String tab = getTab(nbrTab);
        String tabPlus1 = getTab(nbrTab + 1);
        FdbPhysicalFileAccess physical = fileAccess.getPhysicalAccess();
        FdbProperties state = (FdbProperties) physical.readFileObjectAt(FdbSessionImpl.BSTATE_IFA).getData();
        Iterator propIt = state.getKeys().iterator();
        writer.println(tab + "State of base :");
        while (propIt.hasNext()) {
            String key = (String) propIt.next();
            writer.println(tabPlus1 + key + " = " + state.getProperty(key));
        }
        writer.println();
    }

    private Object[] readObjectByFloid(FdbLoid floid) {
        FdbPoid fpoid = oidTable.getFpoid(floid);
        if (fpoid != null) {
            FdbDataObject datas = fileAccess.readDatas(fpoid);
            if (datas != null) {
                return datas.getDatas();
            }
        }
        return null;
    }


    private FdbOidTable oidTable;
    private FdbMetaRepository repository;
    private FdbLogicalSystemPageAccess fileAccess;
    private FdbProperties props;
    private PrintWriter writer;


    private static String propFilePath;
    private static String instanceList;
    private static boolean withMeta = false;
    private static boolean withState = false;
    private static boolean withSelectedInstance = false;
    private static boolean noInstance = false;
    private static final String FILE_OPTION = "-file";
    private static final String INSTANCE_OPTION = "-i";
    private static final String NO_INSTANCE_OPTION = "-ni";
    private static final String META_OPTION = "-m";
    private static final String STATE_OPTION = "-s";
    private static final String HELP_OPTION = "-h";

}
