package org.objectweb.fdb.impl;

import org.objectweb.fdb.FdbLoid;
import org.objectweb.fdb.FdbTransaction;
import org.objectweb.fdb.api.*;
import org.objectweb.fdb.exception.FdbDatabaseException;
import org.objectweb.fdb.exception.FdbIdentityException;
import org.objectweb.fdb.exception.FdbSchemaException;
import org.objectweb.fdb.exception.FdbQueryEngineException;
import org.objectweb.fdb.impl.basic.RecordsFile;
import org.objectweb.fdb.impl.meta.FdbMetaRepository;
import org.objectweb.fdb.impl.meta.ClassMetaDescription;
import org.objectweb.fdb.impl.FdbObjectImpl;
import org.objectweb.fdb.impl.FdbPoid;
import org.objectweb.fdb.impl.FdbProperties;
import org.objectweb.fdb.impl.trace.Trace;

import java.util.Collection;
import java.util.Iterator;

public abstract class FdbSessionImpl implements FdbSessionInternal {

    public FdbSessionImpl(FdbProperties properties) {
        trace = InternalFdbFactory.getTracer(properties);
        sessionId = InternalFdbFactory.getSessionId(properties, this);
        trace.println(Trace.SESSION, "{0} : create new instance with properties : {1}", sessionId, properties);
        cache = InternalFdbFactory.getObjectCache(properties);
        fileAccess = InternalFdbFactory.getLogicalAccess(properties);
        oidTable = InternalFdbFactory.getOidTable(properties);
        oidProvider = InternalFdbFactory.getIdentityProvider(properties);
        repository = InternalFdbFactory.getMetaRepository(this, properties);
        props = properties;
        transaction = createTransaction();
        checkStateInBase();
    }


    public void defineClass(ClassMetaDescription classMetaDescription) {
        trace.println(Trace.SESSION, "{0} : defineClass({1})", sessionId, classMetaDescription);
        repository.defineClass(classMetaDescription);
    }

    public Collection getAllClassNames() {
        return repository.getAllClassNames();
    }

    public boolean isClassDefined(String fullQualifiedClassName) {
        try {
            repository.getClidFromClassName(fullQualifiedClassName);
            return true;
        } catch (FdbSchemaException e) {
        }
        return false;
    }

    public String getClassNameFor(Object oid) {
        trace.println(Trace.SESSION, "{0} : getClassNameFor({1})", sessionId, oid);
        FdbLoid floid = getFloidFromOid(oid);
        if (oidTable.contains(floid)) {
            FdbPoid fpoid = oidTable.getFpoid(floid);
            if (fpoid != null) {
                return repository.getClassNameFromClid(fpoid.getClid());
            }
        }
        return null;
    }

    public Object createObject(Object[] objectToCreate, Class objectClass) {
        //        trace.println(Trace.SESSION, "{0} : createObject({1}, {2})", sessionId, objectToCreate, objectClass);
        return createObject(makeNewFileOid(objectClass), objectToCreate);
    }

    public Object createObject(Object[] objectToCreate, String objectClassName) {
        return createObject(makeNewFileOid(objectClassName), objectToCreate);
    }

    public boolean contains(Object oid) {
        throw new UnsupportedOperationException("not yet implemented...");
    }

    public Collection getExtent(Class theClass) {
        String fullClassName = theClass.getName();
        return getExtent(fullClassName);
    }


    public void setOptimistic() {
    }

    public void setPessimistic() {
    }

    public void begin() {
        trace.println(Trace.SESSION, "{0} : begin()", sessionId);
        fileAccess.start();
    }

    public void commit() {
        trace.println(Trace.SESSION, "{0} : commit()", sessionId);
        fileAccess.commit();
    }

    public void rollback() {
        trace.println(Trace.SESSION, "{0} : rollback()", sessionId);
        fileAccess.rollback();
        oidProvider.getStateFromBase();
        oidTable.checkStateFromBase(props);
    }


    public FdbQueryManager getQueryManager() {
        if (queryManager == null) {
            try {
                queryManager = (FdbQueryManager) Class.forName("org.objectweb.fdb.impl.qe.QueryManagerImpl").newInstance();
                queryManager.init(this);
            } catch (InstantiationException e) {
                throw new FdbQueryEngineException("cannot instanciate query manager", e);
            } catch (IllegalAccessException e) {
                throw new FdbQueryEngineException("cannot initialize query manager", e);
            } catch (ClassNotFoundException e) {
                throw new FdbQueryEngineException("cannot find query manager class", e);
            }
        }
        return queryManager;
    }

    public FdbTransaction currentTransaction() {
        return transaction;
    }

    public FdbLogicalSystemPageAccess getFileAccess() {
        return fileAccess;
    }

    public FdbMetaRepository getMetaRepository() {
        return repository;
    }

    public FdbSessionInternal getInternalSession() {
        return this;
    }

    public FdbProperties getProperties() {
        return props;
    }

    public boolean isNewBase() {
        return fileAccess.getPhysicalAccess().isNewBase();
    }

    public IdentityProviderImpl getOidProvider() {
        return oidProvider;
    }

    public Object getSessionId() {
        return sessionId;
    }

    public void clearObjectCache() {
        trace.println(Trace.SESSION, "{0} : clearObjectCache()", sessionId);
        cache.clear();
    }

    public void restructureBase() {
        trace.println(Trace.SESSION, "{0} : restructureBase()", sessionId);

        if (transaction.isActive()) {
            throw new FdbDatabaseException("Transaction must not be active during restructureBase()");
        }

        try {
            transaction.begin();
            fileAccess.cleanAllNullValue();
            transaction.commit();
        } catch (Exception e) {
            throw new FdbDatabaseException("", e);
        }

        try {
            Iterator it = fileAccess.getPhysicalAccess().getRAFFiles();
            while (it.hasNext()) {
                RecordsFile rf = (RecordsFile) it.next();
                rf.restructure();
            }
        } catch (Exception e) {
            throw new FdbDatabaseException("", e);
        }
    }

    public String toString() {
        return "session " + sessionId;
    }

    /****************************************    PRIVATE    **************************************/

    protected FdbLoid getFloidFromOid(Object oid) {
        //        trace.println(Trace.SESSION, "{0} : getFloidFromOid({1})", sessionId, oid);
        try {
            return (FdbLoid) oid;
        } catch (ClassCastException cce) {
            throw new FdbIdentityException("oid is not a FileLoid but a " + oid.getClass().getName());
        }
    }

    private void checkStateInBase() {
        trace.println(Trace.SESSION, "{0} : checkStateInBase()", sessionId);
        FdbProperties state;
        FdbPhysicalFileAccess physical = fileAccess.getPhysicalAccess();

        try {
            state = (FdbProperties) physical.readFileObjectAt(BSTATE_IFA).getData();
        } catch (FdbDatabaseException e) {
            physical.updateObjectInBase(BSTATE_IFA, props, false);
            return;
        }

        if (!state.getImplementation().equalsIgnoreCase(props.getImplementation())) {
            StringBuffer sb = new StringBuffer();
            sb.append("The implementation of the existing base ").append(props.getDbFileFullName());
            sb.append(" : ").append(state.getImplementation());
            sb.append(" is different of the implementation declared in properties : ");
            sb.append(props.getImplementation());
            throw new FdbDatabaseException(sb.toString());
        }
    }

    protected FdbTransaction createTransaction() {
        return new FdbTransactionImpl(this);
    }




    protected FdbOidTable oidTable;
    protected FdbCache cache;
    protected IdentityProviderImpl oidProvider;
    protected FdbLogicalSystemPageAccess fileAccess;
    protected FdbMetaRepository repository;
    protected FdbQueryManager queryManager;
    protected FdbProperties props;
    protected Object sessionId;
    protected FdbTransaction transaction;
    protected Trace trace;

    public static final InFileAddress BSTATE_IFA = new InFileAddress("bState");
}
