package org.objectweb.fdb.impl;

import org.objectweb.fdb.api.FdbObject;
import org.objectweb.fdb.api.FdbPhysicalFileAccessInternal;
import org.objectweb.fdb.exception.FdbDatabaseException;
import org.objectweb.fdb.impl.FdbProperties;
import org.objectweb.fdb.impl.InFileAddress;

import java.util.*;

public class InTransactionBaseImage {
    private Map deletedObjects;
    private Map modifiedObjects;
    private Map newObjects;
    private Map readObjects;
    private boolean isWorking;
    private FdbPhysicalFileAccessInternal physicalAccess;
    private boolean keepInMemory;


    public InTransactionBaseImage(FdbPhysicalFileAccessInternal physicalAccess, FdbProperties properties) {
        deletedObjects = Collections.synchronizedMap(new HashMap());
        newObjects = Collections.synchronizedMap(new HashMap());
        modifiedObjects = Collections.synchronizedMap(new HashMap());
        readObjects = Collections.synchronizedMap(new HashMap());
        this.physicalAccess = physicalAccess;
        isWorking = false;
        keepInMemory = properties.isKeepingInMemory();
    }

    public boolean isWorking() {
        return isWorking;
    }

    public void start() {
    }

    public synchronized void commit() {
        isWorking = true;
        try {
            synchronized (deletedObjects) {
                Iterator values = deletedObjects.values().iterator();
                while (values.hasNext()) {
                    physicalAccess.internalDeleteFileObject((InFileAddress) values.next());
                }
            }

            synchronized (newObjects) {
                Iterator values = newObjects.values().iterator();
                while (values.hasNext()) {
                    FdbObject fo = (FdbObject) values.next();
                    physicalAccess.internalInsertFileObject(fo.getIfa(), fo);
                }
            }

            synchronized (modifiedObjects) {
                Iterator values = modifiedObjects.values().iterator();
                while (values.hasNext()) {
                    FdbObject fo = (FdbObject) values.next();
                    physicalAccess.internalUpdateFileObject(fo.getIfa(), fo);
                }
            }

            finishTransaction();
        } finally {
            isWorking = false;
        }
    }

    public synchronized void rollback() {
        isWorking = true;
        try {
            finishTransaction();
        } finally {
            isWorking = false;
        }
    }

    public void finishTransaction() {
        if (!keepInMemory) {
                readObjects.clear();
        }
        modifiedObjects.clear();
        newObjects.clear();
        deletedObjects.clear();
    }


    public void insertFileObject(InFileAddress ifa, FdbObject fo) {
        isWorking = true;
        try {
            String address = ifa.getAddress();
            if (modifiedObjects.containsKey(address) ||
                    newObjects.containsKey(address) ||
                    readObjects.containsKey(address)) {
                isWorking = false;
                throw new FdbDatabaseException("you could not insert a object already in base : "+ifa + ", "+fo);
            }
            if (deletedObjects.containsKey(address)) {
                deletedObjects.remove(address);
                modifiedObjects.put(address, fo);
            } else {
                newObjects.put(address, fo);
            }
            readObjects.put(address, fo);
        } finally {
            isWorking = false;
        }
    }

    public FdbObject readFileObjectAt(InFileAddress ifa) {
        isWorking = true;
        try {
            String address = ifa.getAddress();
            if (deletedObjects.containsKey(address)) {
                throw new FdbDatabaseException("object has been removed from base : "+ifa);
            }
            FdbObject result = (FdbObject) readObjects.get(address);
            if (result == null) {
                result = physicalAccess.internalReadFileObjectAt(ifa);
                readObjects.put(address, result);
            }
            return result;
        } finally {
            isWorking = false;
        }

    }

    public void updateFileObject(InFileAddress ifa, FdbObject fo) {
        isWorking = true;
        try {
            String address = ifa.getAddress();
            if (deletedObjects.containsKey(address)) {
                isWorking = false;
                throw new FdbDatabaseException("object has been removed from base : "+ifa + ", "+fo);
            }
            if (newObjects.containsKey(address)) {
                newObjects.put(address, fo);
            } else {
                modifiedObjects.put(address, fo);
            }
            readObjects.put(address, fo);
        } finally {
            isWorking = false;
        }
    }

    public void deleteFileObject(InFileAddress ifa) {
        isWorking = true;
        try {
            String address = ifa.getAddress();
            if (newObjects.containsKey(address)) {
                newObjects.remove(address);
            } else {
                deletedObjects.put(address, ifa);
                modifiedObjects.remove(address);
            }
            readObjects.remove(address);
        } finally {
            isWorking = false;
        }
    }
}
