package org.objectweb.fdb.impl;

import org.objectweb.fdb.FdbSession;
import org.objectweb.fdb.exception.FdbDatabaseException;
import org.objectweb.fdb.api.*;
import org.objectweb.fdb.impl.implempty.GenericCacheImplEmpty;
import org.objectweb.fdb.impl.implempty.PhysicalFileAccessInMemoryImpl;
import org.objectweb.fdb.impl.implpage.FdbSessionPageImpl;
import org.objectweb.fdb.impl.impltx.OidTableTxImpl;
import org.objectweb.fdb.impl.impltx.FdbSessionTxImpl;
import org.objectweb.fdb.impl.lock.LockTableImpl;
import org.objectweb.fdb.impl.meta.FdbMetaRepository;
import org.objectweb.fdb.impl.cache.GenericCache;
import org.objectweb.fdb.impl.DataObjectImpl;
import org.objectweb.fdb.impl.FdbProperties;
import org.objectweb.fdb.impl.IdentityProviderImpl;
import org.objectweb.fdb.impl.trace.Trace;
import org.objectweb.fdb.impl.trace.TraceContext;

import java.util.Hashtable;
import java.util.Map;
import java.io.File;

public class InternalFdbFactory {
    private static Map fileDbProperties = new Hashtable();
    private static Map oidTables = new Hashtable();
    private static Map logicals = new Hashtable();
    private static Map physicals = new Hashtable();
    private static Map metaRepositories = new Hashtable();
    private static Map oidProviders = new Hashtable();
    private static Map lockManagers = new Hashtable();
    private static Map sessionIdMap = new Hashtable();
    private static Map sessions = new Hashtable();
    private static Map tracers = new Hashtable();
    public static final String PAGE_IMPLEMENTATION = "mono";
    public static final String TX_IMPLEMENTATION = "multi";

    public static synchronized Trace getTracer(FdbProperties properties) {
        String path = properties.getDbFileFullName();
        if (!tracers.containsKey(path)) {
            TraceContext traceContext = new TraceContext(properties.isTraceEnable(), properties.getTraceModuleNames());
            Trace trace = traceContext.getTracer();
            tracers.put(path, trace);
        }
        return (Trace) tracers.get(path);
    }

    public static synchronized FdbOidTable getOidTable(FdbProperties properties) {
        String path = properties.getDbFileFullName();
        if (!oidTables.containsKey(path)) {
            FdbOidTable oidTable;
            FdbPhysicalFileAccess physicalAccess = getPhysicalAccess(properties);
            if (properties.getImplementation().equalsIgnoreCase(PAGE_IMPLEMENTATION)) {
                oidTable = OidTableImpl.getAnOidTable(physicalAccess, properties);
            } else {
                oidTable = OidTableTxImpl.getAnOidTable(physicalAccess, properties);
            }
            oidTables.put(path, oidTable);
        }
        return (FdbOidTable) oidTables.get(path);
    }

    public static synchronized FdbMetaRepository getMetaRepository(FdbSessionInternal session,
                                                                       FdbProperties properties) {
        String path = properties.getDbFileFullName();
        if (!metaRepositories.containsKey(path)) {
            FdbMetaRepository metaRepository = FdbMetaRepository.getAnMetaRepository(session, properties);
            metaRepositories.put(path, metaRepository);
        }
        FdbMetaRepository result = (FdbMetaRepository) metaRepositories.get(path);
        result.addSession(session);
        return result;
    }

    public static synchronized FdbLockTable getLockTable(FdbProperties properties) {
        String path = properties.getDbFileFullName();
        if (!lockManagers.containsKey(path)) {
            lockManagers.put(path, new LockTableImpl());
        }
        return (FdbLockTable) lockManagers.get(path);
    }

    public static synchronized IdentityProviderImpl getIdentityProvider(FdbProperties properties) {
        String path = properties.getDbFileFullName();
        if (!oidProviders.containsKey(path)) {
            FdbPhysicalFileAccess physicalAccess = getPhysicalAccess(properties);
            IdentityProviderImpl oidProvider = IdentityProviderImpl.getAnIdentityProvider(physicalAccess);
            oidProviders.put(path, oidProvider);
        }
        return (IdentityProviderImpl) oidProviders.get(path);
    }

    public static synchronized FdbLogicalSystemPageAccess getLogicalAccess(FdbProperties properties) {
        String path = properties.getDbFileFullName();
        if (!logicals.containsKey(path)) {
            FdbLogicalSystemPageAccess logical = new LogicalSystemPageAccessImpl(properties);
            logicals.put(path, logical);
        }
        return (FdbLogicalSystemPageAccess) logicals.get(path);
    }

    public static synchronized FdbPhysicalFileAccess getPhysicalAccess(FdbProperties properties) {
        String path = properties.getDbFileFullName();
        if (!physicals.containsKey(path)) {
            FdbPhysicalFileAccess physical = null;
            String dbPath = properties.getDbFileFullName();
            if (!properties.isInMemoryFile()) {
                File f = new File(dbPath);
                try {
                    if (f.exists()) {
                        physical = new PhysicalFileAccessImpl("rw", properties);
                    } else {
                        physical = new PhysicalFileAccessImpl(properties.getInitialSize(), properties);
                    }
                } catch (Exception e) {
                    throw new FdbDatabaseException("Could not create physical access to base", e);
                }
            } else {
                physical = new PhysicalFileAccessInMemoryImpl(properties);
            }
            physicals.put(path, physical);
        }
        return (FdbPhysicalFileAccess) physicals.get(path);
    }

    public static FdbDataObject getNewDataObject(FdbProperties properties, Object[] arrayOfObjects) {
        return DataObjectImpl.newInstance(arrayOfObjects);
    }

    public static synchronized FdbSession getNewSession(String propertiesFilePath) {
        FdbProperties props = getProperties(propertiesFilePath);
        FdbSession session;
        if (props.getImplementation().equalsIgnoreCase(TX_IMPLEMENTATION)) {
            session = new FdbSessionTxImpl(props);
        } else {
            session = new FdbSessionPageImpl(props);
        }
        // initialise stuff
        session.currentTransaction().begin();
        session.currentTransaction().commit();
        return session;
    }

    // each session has his own cache
    public static synchronized FdbCache getObjectCache(FdbProperties properties) {
        FdbCache cache;
        int size = Integer.parseInt(properties.getProperty(FdbProperties.OBJECT_CACHE_SIZE_KEY));
        if (size > 0) {
            cache = new GenericCache(size, "objectCache",
                                     properties.getCacheClearPourcent(),
                                     getTracer(properties));
        } else {
            cache = new GenericCacheImplEmpty("Run with empty object cache...");
        }
        return cache;
    }

    public static synchronized FdbProperties getProperties(String path) {
        if (!fileDbProperties.containsKey(path)) {
            FdbProperties p = new FdbProperties(path);
            fileDbProperties.put(path, p);
        }
        return (FdbProperties) fileDbProperties.get(path);
    }

    public static Object getSessionById(Object sessionId) {
        return sessions.get(sessionId);
    }

    public static synchronized Object getSessionId(FdbProperties properties, FdbSession session) {
        String path = properties.getDbFileFullName();
        if (!sessionIdMap.containsKey(path)) {
            Object id = new Integer(0);
            sessionIdMap.put(path, id);
            sessions.put(id, session);
            return id;
        }
        Integer oldId = (Integer) sessionIdMap.get(path);
        Integer newId = new Integer(oldId.intValue() + 1);
        sessionIdMap.put(path, newId);
        sessions.put(newId, session);
        return newId;
    }
}
