package org.objectweb.fdb.impl;

import org.objectweb.fdb.api.*;
import org.objectweb.fdb.impl.page.*;
import org.objectweb.fdb.impl.*;
import org.objectweb.fdb.impl.trace.Trace;

import java.util.*;

public class LogicalSystemPageAccessImpl implements FdbLogicalSystemPageAccess {

    public LogicalSystemPageAccessImpl(FdbProperties properties) {
        tracer = InternalFdbFactory.getTracer(properties);
        tracer.println(Trace.LOGICAL, "LogicalSystemPageAccessImpl : create new instance");
        physicalAccess = InternalFdbFactory.getPhysicalAccess(properties);
        ifaIndex = IfaIndexImpl.getAnIfaIndex(physicalAccess, properties);
        oidProvider = InternalFdbFactory.getIdentityProvider(properties);
        instancePageSize = properties.getInstancePageSize();
        systemPageSize = properties.getSystemPageSize();
        instancePageNumber = properties.getInstanceDbFileNumber();
        getInMemoryInfos();
    }

    public String toString() {
        return physicalAccess.toString();
    }

    public FdbPhysicalFileAccess getPhysicalAccess() {
        return physicalAccess;
    }

    public void start() {
        tracer.println(Trace.LOGICAL, "LogicalSystemPageAccessImpl : start() : nothing to do");
    }

    public void commit() {
        tracer.println(Trace.LOGICAL, "LogicalSystemPageAccessImpl : commit()");
        physicalAccess.commit();
    }

    public void rollback() {
        tracer.println(Trace.LOGICAL, "LogicalSystemPageAccessImpl : rollback()");
        physicalAccess.rollback();
        ifaIndex.getStateFromBase();
    }

    /***********************************  INSTANCE  **********************************************/

    public FdbPoid allocateNewFpoid(Object clid) {
        tracer.println(Trace.LOGICAL, "LogicalSystemPageAccessImpl : allocateNewFpoid({0})", clid);
        synchronized (instanceFreeSpaces) {
            // find the next instance page to use
            InFileAddress availableSystemIfa = getFreeSystemIfaIn(instanceFreeSpaces);
            PageInfo infos = null;
            SystemPage systemPage = null;
            if (availableSystemIfa != null) {
                systemPage = (SystemPage) physicalAccess.readFileObjectAt(availableSystemIfa);
                infos = (PageInfo) systemPage.getDataAt(availableSystemIfa);

            } else {
                systemPage = getOrCreateAvailableSystemPage();
                infos = buildNewInstancePage(oidProvider.getNewPid(), systemPage.getIfa());
                systemPage.setDataAt(systemPage.getIfa(), infos);
            }
            Object iid = infos.getFreePlace();
            ((InMemoryPageInfo) instanceFreeSpaces.get(infos.getPid())).setFreeSpace(infos.getEmptySpace());
            physicalAccess.updateFileObject(systemPage);
            return new FdbPoid(clid, infos.getPid(), iid);
        }
    }

    public void desallocateFpoid(FdbPoid fpoid) {
        tracer.println(Trace.LOGICAL, "LogicalSystemPageAccessImpl : desallocateFpoid({0})", fpoid);
        deleteDatas(fpoid);
    }

    // allocation is done
    public void insertDatas(FdbPoid fpoid, FdbDataObject datas) {
        tracer.println(Trace.LOGICAL, "LogicalSystemPageAccessImpl : insertDatas({0}, {1})", fpoid, datas);
        InFileAddress instanceIfa = getInstanceIfa(fpoid);
        InstancePage dataPage = (InstancePage) physicalAccess.readFileObjectAt(instanceIfa);
        dataPage.setDataAt(instanceIfa, datas);
        physicalAccess.updateFileObject(dataPage);
    }

    public FdbDataObject readDatas(FdbPoid fpoid) {
        tracer.println(Trace.LOGICAL, "LogicalSystemPageAccessImpl : readDatas({0})", fpoid);
        InFileAddress ifa = getInstanceIfa(fpoid);
        InstancePage dataPage = (InstancePage) physicalAccess.readFileObjectAt(ifa);
        return (FdbDataObject) dataPage.getDataAt(ifa);
    }

    public FdbPoid updateDatas(FdbPoid fpoid, FdbDataObject datas) {
        tracer.println(Trace.LOGICAL, "LogicalSystemPageAccessImpl : updateDatas({0}, {1})", fpoid, datas);
        InFileAddress ifa = getInstanceIfa(fpoid);
        InstancePage dataPage = (InstancePage) physicalAccess.readFileObjectAt(ifa);
        dataPage.setDataAt(ifa, datas);
        physicalAccess.updateFileObject(dataPage);
        return fpoid;
    }

    public void deleteDatas(FdbPoid fpoid) {
        tracer.println(Trace.LOGICAL, "LogicalSystemPageAccessImpl : deleteDatas({0})", fpoid);
        // get ifa of the system page
        InFileAddress systemIfa = ifaIndex.getPidIfa(fpoid.getPid());
        // get the page
        SystemPage systemPage = (SystemPage) physicalAccess.readFileObjectAt(systemIfa);
        synchronized (instanceFreeSpaces) {
            // get the page info
            PageInfo infos = (PageInfo) systemPage.getDataAt(systemIfa);
            infos.freeSpaceAt(fpoid.getIidAsInteger());
            ((InMemoryPageInfo) instanceFreeSpaces.get(fpoid.getPid())).setFreeSpace(infos.getEmptySpace());
            // update page where modifications has been done
            physicalAccess.updateFileObject(systemPage);
        }
    }


    public void cleanAllNullValue() {
        tracer.println(Trace.LOGICAL, "LogicalSystemPageAccessImpl : cleanAllNullValue()");
        Iterator pidIfas = ifaIndex.getAllPidIfa().iterator();
        while (pidIfas.hasNext()) {
            InFileAddress ifa = (InFileAddress) pidIfas.next();
            SystemPage systemPage = (SystemPage) physicalAccess.readFileObjectAt(ifa);
            PageInfo pageInfo = (PageInfo) systemPage.getDataAt(ifa);
            InFileAddress instanceIfa = pageInfo.getPageIfa();
            Page dataPage = (Page) physicalAccess.readFileObjectAt(instanceIfa);
            Iterator freePlaces = pageInfo.getAllFreePlace();
            while (freePlaces.hasNext()) {
                instanceIfa.setIndex(((Integer) freePlaces.next()).intValue());
                dataPage.setDataAt(instanceIfa, null);
            }
            physicalAccess.updateFileObject(dataPage);
        }
    }

    /***************************************  INDEX  *************************************************/

    public InFileAddress allocateNodeAddressAndInsert(FdbObject node) {
        tracer.println(Trace.LOGICAL, "LogicalSystemPageAccessImpl : allocateNodeAddressAndInsert()");
        synchronized (nodeFreeSpaces) {
            InFileAddress availableSystemIfa = getFreeSystemIfaIn(nodeFreeSpaces);
            PageInfo infos = null;
            SystemPage systemPage = null;
            if (availableSystemIfa != null) {
                availableSystemIfa = availableSystemIfa.getClone();
                systemPage = (SystemPage) physicalAccess.readFileObjectAt(availableSystemIfa);
                infos = (PageInfo) systemPage.getDataAt(availableSystemIfa);
            } else {
                systemPage = getOrCreateAvailableSystemPage();
                infos = buildNewNodePage(oidProvider.getNewPid(), systemPage.getIfa());
                systemPage.setDataAt(systemPage.getIfa(), infos);
            }
            Integer iid = infos.getFreePlace();
            ((InMemoryPageInfo) nodeFreeSpaces.get(infos.getPid())).setFreeSpace(infos.getEmptySpace());
            InFileAddress pageIfa = infos.getPageIfa();
            pageIfa.setIndex(iid.intValue());
            node.setIfa(pageIfa);
            Page nodePage = (Page) physicalAccess.readFileObjectAt(pageIfa);
            nodePage.setDataAt(pageIfa, node);
            physicalAccess.updateFileObject(systemPage);
            physicalAccess.updateFileObject(nodePage);
            return pageIfa.getClone();
        }
    }

    public Object readNode(InFileAddress ifa) {
        Page page = (Page) physicalAccess.readFileObjectAt(ifa);
        return page.getDataAt(ifa);
    }

    public void updateNode(InFileAddress ifa, FdbObject node) {
        tracer.println(Trace.LOGICAL, "LogicalSystemPageAccessImpl : updateNode({0}, {1})", ifa, node);
        Page page = (Page) physicalAccess.readFileObjectAt(ifa);
        page.setDataAt(ifa, node);
        physicalAccess.updateFileObject(page);
    }

    public void removeNode(InFileAddress ifa) {
        tracer.println(Trace.LOGICAL, "LogicalSystemPageAccessImpl : removeNode({0})", ifa);
        NodePage nodePage = (NodePage) physicalAccess.readFileObjectAt(ifa);
        InFileAddress systemIfa = ifaIndex.getPidIfa(nodePage.getPid());
        SystemPage systemPage = (SystemPage) physicalAccess.readFileObjectAt(systemIfa);
        synchronized (nodeFreeSpaces) {
            PageInfo infos = (PageInfo) systemPage.getDataAt(systemIfa);
            infos.freeSpaceAt(new Integer(ifa.getIndex()));
            ((InMemoryPageInfo) nodeFreeSpaces.get(nodePage.getPid())).setFreeSpace(infos.getEmptySpace());
            physicalAccess.updateFileObject(systemPage);
        }
    }

    public InFileAddress allocateLeafAddressAndInsert(FdbObject oids) {
        tracer.println(Trace.LOGICAL, "LogicalSystemPageAccessImpl : allocateLeafAddressAndInsert({0})");
        synchronized (leafFreeSpaces) {
            InFileAddress availableSystemIfa = getFreeSystemIfaIn(leafFreeSpaces);
            PageInfo infos = null;
            SystemPage systemPage = null;
            if (availableSystemIfa != null) {
                availableSystemIfa = availableSystemIfa.getClone();
                systemPage = (SystemPage) physicalAccess.readFileObjectAt(availableSystemIfa);
                infos = (PageInfo) systemPage.getDataAt(availableSystemIfa);
            } else {
                systemPage = getOrCreateAvailableSystemPage();
                infos = buildNewLeafPage(oidProvider.getNewPid(), systemPage.getIfa());
                systemPage.setDataAt(systemPage.getIfa(), infos);
            }
            Integer iid = infos.getFreePlace();
            ((InMemoryPageInfo) leafFreeSpaces.get(infos.getPid())).setFreeSpace(infos.getEmptySpace());
            InFileAddress pageIfa = infos.getPageIfa();
            pageIfa.setIndex(iid.intValue());
            oids.setIfa(pageIfa);
            Page nodePage = (Page) physicalAccess.readFileObjectAt(pageIfa);
            nodePage.setDataAt(pageIfa, oids);
            physicalAccess.updateFileObject(systemPage);
            physicalAccess.updateFileObject(nodePage);
            return pageIfa.getClone();
        }
    }

    public Object readLeaf(InFileAddress ifa) {
        Page page = (Page) physicalAccess.readFileObjectAt(ifa);
        return page.getDataAt(ifa);
    }

    public void updateLeaf(InFileAddress ifa, FdbObject oids) {
        updateNode(ifa, oids);
    }

    public void removeLeaf(InFileAddress ifa) {
        tracer.println(Trace.LOGICAL, "LogicalSystemPageAccessImpl : removeNode({0})", ifa);
        LeafPage leafPage = (LeafPage) physicalAccess.readFileObjectAt(ifa);
        InFileAddress systemIfa = ifaIndex.getPidIfa(leafPage.getPid());
        SystemPage systemPage = (SystemPage) physicalAccess.readFileObjectAt(systemIfa);
        synchronized (leafFreeSpaces) {
            PageInfo infos = (PageInfo) systemPage.getDataAt(systemIfa);
            infos.freeSpaceAt(new Integer(ifa.getIndex()));
            ((InMemoryPageInfo) leafFreeSpaces.get(leafPage.getPid())).setFreeSpace(infos.getEmptySpace());
            physicalAccess.updateFileObject(systemPage);
        }
    }

    /***********************************     PRIVATE     *********************************************/

    private InFileAddress getInstanceIfa(FdbPoid fpoid) {
        InFileAddress systemIfa = ifaIndex.getPidIfa(fpoid.getPid());
        SystemPage systemPage = (SystemPage) physicalAccess.readFileObjectAt(systemIfa);
        PageInfo infos = (PageInfo) systemPage.getDataAt(systemIfa);
        InFileAddress instanceIfa = infos.getPageIfa();
        instanceIfa.setIndex(fpoid.getIidAsInteger().intValue());
        return instanceIfa;
    }

    private void getInMemoryInfos() {
        instanceFreeSpaces = Collections.synchronizedMap(new HashMap());
        leafFreeSpaces = Collections.synchronizedMap(new HashMap());
        nodeFreeSpaces = Collections.synchronizedMap(new HashMap());

        Iterator pids = ifaIndex.getAllPid().iterator();
        while (pids.hasNext()) {
            Object pid = pids.next();
            InFileAddress ifa = ifaIndex.getPidIfa(pid);
            SystemPage systemPage = (SystemPage) physicalAccess.readFileObjectAt(ifa);
            PageInfo pageInfo = (PageInfo) systemPage.getDataAt(ifa);
            InMemoryPageInfo memoryPageInfo = new InMemoryPageInfo(pageInfo);
            if (pageInfo.getPageType() == Page.INSTANCEPAGE_TYPE) {
                instanceFreeSpaces.put(pid, memoryPageInfo);
            } else if (pageInfo.getPageType() == Page.LEAFPAGE_TYPE) {
                leafFreeSpaces.put(pid, memoryPageInfo);
            } else {
                nodeFreeSpaces.put(pid, memoryPageInfo);
            }
        }
    }

    private int getInstanceFileIndex() {
        fileIndex = (fileIndex + 1) % instancePageNumber;
        return (fileIndex + 2);
    }

    private int getLeafFileIndex() {
        fileIndex = (fileIndex + 1) % instancePageNumber;
        return (fileIndex + 2);
    }

    private int getNodeFileIndex() {
        fileIndex = (fileIndex + 1) % instancePageNumber;
        return (fileIndex + 2);
    }

    private synchronized InFileAddress getFreeSystemIfaIn(Map infos) {
        synchronized (infos) {
            if (!infos.isEmpty()) {
                Iterator pids = infos.keySet().iterator();
                while (pids.hasNext()) {
                    Object pid = pids.next();
                    InMemoryPageInfo memoryInfo = (InMemoryPageInfo) infos.get(pid);
                    if (memoryInfo.getFreeSpace() > 0) {
                        return ifaIndex.getPidIfa(pid);
                    }
                }
            }
        }
        return null;
    }

    private synchronized PageInfo buildNewInstancePage(Object pid, InFileAddress systemIfa) {
        InstancePage page = new InstancePage(instancePageSize);
        InFileAddress pageIfa = new InFileAddress(ifaIndex.getNewInstancePageAddress());
        pageIfa.setFileIndex(getInstanceFileIndex());
        page.setIfa(pageIfa);
        physicalAccess.insertFileObject(page);

        PageInfo pageInfo = new PageInfo(instancePageSize, Page.INSTANCEPAGE_TYPE);
        pageInfo.setPageIfa(pageIfa);
        pageInfo.setPid(pid);

        instanceFreeSpaces.put(pid, new InMemoryPageInfo(pageInfo));
        ifaIndex.putPidIfa(pid, systemIfa.getClone());

        return pageInfo;
    }

    private synchronized PageInfo buildNewLeafPage(Object pid, InFileAddress systemIfa) {
        LeafPage page = new LeafPage(leafPageSize, pid);
        InFileAddress pageIfa = new InFileAddress(ifaIndex.getNewLeafPageAddress());
        pageIfa.setFileIndex(getLeafFileIndex());
        page.setIfa(pageIfa);
        physicalAccess.insertFileObject(page);

        PageInfo pageInfo = new PageInfo(leafPageSize, Page.LEAFPAGE_TYPE);
        pageInfo.setPageIfa(pageIfa);
        pageInfo.setPid(pid);

        leafFreeSpaces.put(pid, new InMemoryPageInfo(pageInfo));
        ifaIndex.putPidIfa(pid, systemIfa.getClone());

        return pageInfo;
    }

    private synchronized PageInfo buildNewNodePage(Object pid, InFileAddress systemIfa) {
        NodePage page = new NodePage(nodePageSize, pid);
        InFileAddress pageIfa = new InFileAddress(ifaIndex.getNewNodePageAddress());
        pageIfa.setFileIndex(getNodeFileIndex());
        page.setIfa(pageIfa);
        physicalAccess.insertFileObject(page);

        PageInfo pageInfo = new PageInfo(nodePageSize, Page.NODEPAGE_TYPE);
        pageInfo.setPageIfa(pageIfa);
        pageInfo.setPid(pid);

        nodeFreeSpaces.put(pid, new InMemoryPageInfo(pageInfo));
        ifaIndex.putPidIfa(pid, systemIfa.getClone());

        return pageInfo;
    }

    private synchronized SystemPage getOrCreateAvailableSystemPage() {
        InFileAddress availableSystemPageIfa = ifaIndex.getCurrentAvalaibleSystemPage().getClone();
        SystemPage systemPage = null;
        if (availableSystemPageIfa.getIndex() == (systemPageSize - 1)) {
            availableSystemPageIfa = new InFileAddress(ifaIndex.getNewSystemPageAddress());
            systemPage = new SystemPage(systemPageSize);
            systemPage.setIfa(availableSystemPageIfa);
            physicalAccess.insertFileObject(systemPage);
        } else {
            availableSystemPageIfa.setIndex(availableSystemPageIfa.getIndex() + 1);
            systemPage = (SystemPage) physicalAccess.readFileObjectAt(availableSystemPageIfa);
            systemPage.setIfa(availableSystemPageIfa);
        }
        ifaIndex.setCurrentAvalaibleSystemPage(availableSystemPageIfa);
        return systemPage;
    }


    private IfaIndexImpl ifaIndex;
    private IdentityProviderImpl oidProvider;
    private FdbPhysicalFileAccess physicalAccess;

    private Map instanceFreeSpaces;
    private Map leafFreeSpaces;
    private Map nodeFreeSpaces;

    private int fileIndex = 0;
    private int instancePageNumber;
    private short instancePageSize;
    private short systemPageSize;
    private short leafPageSize = LeafPage.SIZE;
    private short nodePageSize = NodePage.SIZE;
    private Trace tracer;
}
