package org.objectweb.fdb.impl;

import org.objectweb.fdb.FdbLoid;
import org.objectweb.fdb.api.FdbOidTable;
import org.objectweb.fdb.api.FdbObject;
import org.objectweb.fdb.api.FdbPhysicalFileAccess;
import org.objectweb.fdb.exception.FdbDatabaseException;
import org.objectweb.fdb.impl.page.OidPage;
import org.objectweb.fdb.impl.cache.GenericCache;
import org.objectweb.fdb.impl.*;
import org.objectweb.fdb.impl.trace.Trace;

import java.util.*;

public class OidTableImpl implements FdbOidTable {

    protected OidTableImpl(FdbPhysicalFileAccess physicalAccess, FdbProperties properties) {
        tracer = InternalFdbFactory.getTracer(properties);
        atSize = properties.getOidTableSize();
        tracer.println(Trace.OIDTABLE, "OidTableImplPage : create new instance with size = {0}", new Integer(atSize));

        if (atSize > 0) {
            floidTable = new GenericCache(atSize, "atCache",
                                          properties.getCacheClearPourcent(),
                                          InternalFdbFactory.getTracer(properties));
        } else {
            floidTable = new HashMap();
        }
        spaceTable = new HashMap();
        this.physicalAccess = physicalAccess;
    }

    public synchronized void insertFpoid(FdbLoid floid, FdbPoid fpoid) {
        tracer.println(Trace.OIDTABLE, "OidTableImpl : insertFpoid({0}, {1})", floid, fpoid);
        OidInfo oidInfo = new OidInfo(floid, fpoid);
        floidTable.put(floid, oidInfo);
        updateOidInfoInBase(oidInfo, true);
    }

    public synchronized FdbPoid updatePoid(FdbLoid floid, FdbPoid newFpoid) {
        tracer.println(Trace.OIDTABLE, "OidTableImpl : updatePoid({0}, {1})", floid, newFpoid);
        OidInfo info = (OidInfo) floidTable.get(floid);
        info.setFpoid(newFpoid);
        updateOidInfoInBase(info, false);
        return newFpoid;
    }

    public synchronized void removeFloid(FdbLoid floid) {
        tracer.println(Trace.OIDTABLE, "OidTableImpl : removeFloid({0})", floid);
        OidInfo info = (OidInfo) floidTable.get(floid);
        floidTable.remove(floid);
        // fpoid at null in this call : we want to remove the info
        info.setFpoid(null);
        updateOidInfoInBase(info, false);
    }


    public synchronized Collection getAllFloidFromClid(Object clid) {
        tracer.println(Trace.OIDTABLE, "OidTableImpl : getAllFloidFromClid({0})", clid);
        Collection c = new ArrayList();

        for (int i = 0; i < oidIfaList.size(); i++) {
            InFileAddress ifa = new InFileAddress((String) oidIfaList.get(i));
            ifa.setFileIndex(FdbProperties.OID_INDEX);

            OidPage oidPage = (OidPage) physicalAccess.readFileObjectAt(ifa);
            Object[] infos = oidPage.getAllDatas();

            for (int j = 0; j < infos.length; j++) {
                OidInfo info = (OidInfo) infos[j];
                if (info != null) {
                    if (info.getFpoid().hasClidEqualsWith(clid)) {
                        c.add(info.getFloid());
                    }
                }
            }
        }
        return c;
    }

    public synchronized FdbPoid getFpoid(FdbLoid floid) {
        tracer.println(Trace.OIDTABLE, "OidTableImplPage : getFpoid({0})", floid);
        if (!floidTable.containsKey(floid)) {
            if (!loadOidPage(floid)) {
                return null;
            }
        }
        OidInfo oidInfo = (OidInfo) floidTable.get(floid);
        if (oidInfo != null) {
            return oidInfo.getFpoid();
        } else {
            return null;
        }
    }

    public synchronized boolean contains(FdbLoid floid) {
        boolean result = floidTable.containsKey(floid);
        if (!result) {
            if (loadOidPage(floid)) {
                result = floidTable.containsKey(floid);
            }
        }
        return result;
    }

    public synchronized void setUpdate(FdbLoid floid, short value) {
        throw new UnsupportedOperationException("should not be used in this implementation...");
    }

    public synchronized short getUpdate(FdbLoid floid) {
        throw new UnsupportedOperationException("should not be used in this implementation...");
    }

    public synchronized void checkStateFromBase(FdbProperties properties) {
        tracer.println(Trace.OIDTABLE, "OidTableImplPage : getStateFromBase()");
        if (atSize > 0) {
            floidTable = new GenericCache(atSize, "atCache",
                                          properties.getCacheClearPourcent(),
                                          InternalFdbFactory.getTracer(properties));
        } else {
            floidTable = new HashMap();
        }
        spaceTable = new HashMap();
        oidIfaList = (List) physicalAccess.readFileObjectAt(OID_IFA_LIST).getData();
        readAtTable();
    }


    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("OidTableImplBasic : \n");
        sb.append("\tfloid list : ").append(floidTable).append("\n");

        return sb.toString();
    }


    /************************************    protected    ***********************************************/

    protected synchronized void createAtList() {
        oidIfaList = new ArrayList();
        updateAtListInBase(false);
    }

    protected synchronized boolean loadOidPage(FdbLoid floid) {
        tracer.println(Trace.OIDTABLE, "OidTableImplPage : loadOidPage({0})", floid);

        while(physicalAccess.getBaseImage().isWorking()) {
            Thread.yield();
        }

        synchronized (physicalAccess.getBaseImage()) {
            int index = floid.getFloidAsInt() % oidPageSize;
            int pageNumber = floid.getFloidAsInt() / oidPageSize;
            InFileAddress atIfa = new InFileAddress(IFA_PREFIXE + String.valueOf(pageNumber));
            atIfa.setIndex(index);
            atIfa.setFileIndex(FdbProperties.OID_INDEX);
            try {
                // could raise exception if oidPage don't exist
                OidPage oidPage = (OidPage) physicalAccess.readFileObjectAt(atIfa);
                Object[] infos = oidPage.getAllDatas();
                for (int j = 0; j < infos.length; j++) {
                    OidInfo info = (OidInfo) infos[j];
                    if (info != null) {
                        floidTable.put(info.getFloid(), info);
                    }
                }
                return true;
            } catch (Exception e) {
                tracer.println(Trace.OIDTABLE, "OidTableImplPage : loadOidPage : oidPage doesn't exist");
            }
            return false;
        }
    }

    protected synchronized void readAtTable() {
        tracer.println(Trace.OIDTABLE, "OidTableImplPage : readAtTable()");

        while(physicalAccess.getBaseImage().isWorking()) {
            Thread.yield();
        }

        synchronized (physicalAccess.getBaseImage()) {
            int dataRead = 0;
            for (int i = 0; i < oidIfaList.size(); i++) {
                InFileAddress ifa = new InFileAddress((String) oidIfaList.get(i));
                ifa.setFileIndex(FdbProperties.OID_INDEX);
                OidPage oidPage = (OidPage) physicalAccess.readFileObjectAt(ifa);
                Object[] infos = oidPage.getAllDatas();
                for (int j = 0; j < infos.length; j++) {
                    OidInfo info = (OidInfo) infos[j];
                    if (info != null) {
                        floidTable.put(info.getFloid(), info);
                        Object o = spaceTable.get(ifa.getAddress());
                        int usedSpace = 0;
                        if (o != null) {
                            usedSpace = ((Integer) o).intValue();
                        }
                        spaceTable.put(ifa.getAddress(), new Integer(usedSpace + 1));
                        dataRead++;
                        if (dataRead == atSize) {
                            return;
                        }
                    }
                }
            }
        }
    }

    private synchronized void updateAtListInBase(boolean isUpdate) {
        tracer.println(Trace.OIDTABLE, "OidTableImpl : readAtTable({0})", new Boolean(isUpdate));

        while(physicalAccess.getBaseImage().isWorking()) {
            Thread.yield();
        }

        synchronized (physicalAccess.getBaseImage()) {
            physicalAccess.updateObjectInBase(OID_IFA_LIST, oidIfaList, isUpdate);
        }
    }

    protected synchronized void updateOidInfoInBase(OidInfo info, boolean isNewData) {
        tracer.println(Trace.OIDTABLE, "OidTableImpl : updateOidInfoInBase({0}, {1})", info, new Boolean(isNewData));

        while(physicalAccess.getBaseImage().isWorking()) {
            Thread.yield();
        }

        synchronized (physicalAccess.getBaseImage()) {
            int index = info.getFloid().getFloidAsInt() % oidPageSize;
            int pageNumber = info.getFloid().getFloidAsInt() / oidPageSize;
            InFileAddress oidIfa = new InFileAddress(IFA_PREFIXE + String.valueOf(pageNumber));
            oidIfa.setIndex(index);
            oidIfa.setFileIndex(FdbProperties.OID_INDEX);

            OidPage oidPage = null;
            try {
                // could raise exception if oidPage don't exist
                oidPage = (OidPage) physicalAccess.readFileObjectAt(oidIfa);
            } catch (FdbDatabaseException fde) {
                if (fde.getMessage().startsWith("could not read object at ifa("+IFA_PREFIXE)) {
                    // create a new OidPage. Don't happends if we are deleting
                    oidPage = createOidPage(oidIfa);
                } else {
                    throw fde;
                }
            }

            int usedSpace = -1;
            try {
                usedSpace = ((Integer) spaceTable.get(oidIfa.getAddress())).intValue();
            } catch (NullPointerException npe) {
                throw new FdbDatabaseException("Error with OidTable : Your base is probably corrupted...");
            }

            if (info.isDeleting()) {
                oidPage.setDataAt(oidIfa, null);
                spaceTable.put(oidIfa.getAddress(), new Integer(usedSpace - 1));
                if (usedSpace == 1) { // i.e. (usedSpace - 1) == 0
                    physicalAccess.deleteFileObject(oidIfa);
    //                System.out.println(atIfa);
                    oidIfaList.remove(oidIfa.getAddress());
                    spaceTable.remove(oidIfa.getAddress());
                    updateAtListInBase(true);
                } else {
                    physicalAccess.updateFileObject(oidPage);
                }
            } else {
                oidPage.setDataAt(oidIfa, info);
                if (isNewData) {
                    spaceTable.put(oidIfa.getAddress(), new Integer(usedSpace + 1));
                }
                physicalAccess.updateFileObject(oidPage);
            }
        }
    }

    private synchronized OidPage createOidPage(InFileAddress atIfa) {
        OidPage oidPage = new OidPage(oidPageSize);
        oidIfaList.add(atIfa.getAddress());
//        System.out.println(Thread.currentThread().toString()+" create OidPage "+atIfa);
        oidPage.setIfa(atIfa);
        spaceTable.put(atIfa.getAddress(), new Integer(0));
        physicalAccess.insertFileObject(oidPage);
        updateAtListInBase(true);
        return oidPage;
    }


    protected FdbPhysicalFileAccess physicalAccess;
    protected List oidIfaList;
    protected Map floidTable;
    protected Map spaceTable;
    protected int oidPageSize;
    protected int atSize;
    protected Trace tracer;


    /***************************************    STATIC    **********************************************/

    public static FdbOidTable getAnOidTable(FdbPhysicalFileAccess physicalAccess, FdbProperties properties) {
        OidTableImpl oidTable;
        try {
            List list = (List) physicalAccess.readFileObjectAt(OID_IFA_LIST).getData();
            oidTable = new OidTableImpl(physicalAccess, properties);
            oidTable.oidIfaList = list;
            oidTable.readAtTable();
        } catch (Exception e) {
            oidTable = new OidTableImpl(physicalAccess, properties);
            oidTable.createAtList();
        }
        oidTable.oidPageSize = properties.getOidPageSize();
        return oidTable;
    }

    protected static final String IFA_PREFIXE = "id";
    protected static final InFileAddress OID_IFA_LIST = new InFileAddress("idlist");
}
