package org.objectweb.fdb.impl;

import org.objectweb.fdb.api.FdbObject;
import org.objectweb.fdb.api.FdbPhysicalFileAccess;
import org.objectweb.fdb.api.FdbPhysicalFileAccessInternal;
import org.objectweb.fdb.exception.FdbDatabaseException;
import org.objectweb.fdb.impl.basic.RecordWriter;
import org.objectweb.fdb.impl.basic.RecordsFile;
import org.objectweb.fdb.impl.basic.RecordsFileException;
import org.objectweb.fdb.impl.basic.RecordsFileInspector;
import org.objectweb.fdb.impl.FdbProperties;
import org.objectweb.fdb.impl.InFileAddress;
import org.objectweb.fdb.impl.InternalFdbFactory;
import org.objectweb.fdb.impl.InTransactionBaseImage;
import org.objectweb.fdb.impl.trace.Trace;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

public class PhysicalFileAccessImpl implements FdbPhysicalFileAccess, FdbPhysicalFileAccessInternal {

    public PhysicalFileAccessImpl(int initialSize, FdbProperties properties)
            throws IOException, RecordsFileException {

        tracer = InternalFdbFactory.getTracer(properties);
        baseImage = new InTransactionBaseImage(this, properties);

        int nbrDbFiles = properties.getInstanceDbFileNumber();
        if (nbrDbFiles > 1) {
            files = new RecordsFile[nbrDbFiles + 2];
            files[0] = new RecordsFile(properties.getSystemDbFilePath(),
                                       initialSize, properties.getKeyLength(),
                                       properties.getAdditionalSpace());
            files[1] = new RecordsFile(properties.getAtDbFilePath(),
                                       initialSize, properties.getKeyLength(),
                                       properties.getAdditionalSpace());

            for (int i = 0; i < nbrDbFiles; i++) {
                files[i + 2] = new RecordsFile(properties.getInstanceDbFilePathAt(i),
                                               initialSize,
                                               properties.getKeyLength(),
                                               properties.getAdditionalSpace());
            }
        } else {
            files = new RecordsFile[1];
            files[0] = new RecordsFile(properties.getSystemDbFilePath(),
                                       initialSize, properties.getKeyLength(),
                                       properties.getAdditionalSpace());
        }

        for (int i = 0; i < files.length; i++) {
            files[i].setTracer(tracer);
        }

        isNewBase = false;

        tracer.println(Trace.PHYSICAL, "new instance with {0} underlying files", new Integer(nbrDbFiles));
    }

    public PhysicalFileAccessImpl(String accessFlags, FdbProperties properties)
            throws IOException, RecordsFileException {

        tracer = InternalFdbFactory.getTracer(properties);
        baseImage = new InTransactionBaseImage(this, properties);

        int nbrDbFiles = properties.getInstanceDbFileNumber();
        if (nbrDbFiles > 1) {
            files = new RecordsFile[nbrDbFiles + 2];
            files[0] = new RecordsFile(properties.getSystemDbFilePath(),
                                       accessFlags, properties.getKeyLength(),
                                       properties.getAdditionalSpace());
            files[1] = new RecordsFile(properties.getAtDbFilePath(),
                                       accessFlags, properties.getKeyLength(),
                                       properties.getAdditionalSpace());

            for (int i = 0; i < nbrDbFiles; i++) {
                files[i + 2] = new RecordsFile(properties.getInstanceDbFilePathAt(i),
                                               accessFlags,
                                               properties.getKeyLength(),
                                               properties.getAdditionalSpace());
            }
        } else {
            files = new RecordsFile[1];
            files[0] = new RecordsFile(properties.getSystemDbFilePath(),
                                       accessFlags, properties.getKeyLength(),
                                       properties.getAdditionalSpace());
        }

        for (int i = 0; i < files.length; i++) {
            files[i].setTracer(tracer);
        }

        isNewBase = true;

        tracer.println(Trace.PHYSICAL, "new instance with {0} underlying files", new Integer(nbrDbFiles));
    }

    public boolean isNewBase() {
        return isNewBase;
    }


    public void insertFileObject(FdbObject fo) {
        tracer.println(Trace.PHYSICAL, "insertFileObject({0})", fo);
        try {
            baseImage.insertFileObject(fo.getIfa(), fo);
        } catch (Exception e) {
            throw new FdbDatabaseException("could not insert object " + fo, e);
        }
    }

    public FdbObject readFileObjectAt(InFileAddress ifa) {
        tracer.println(Trace.PHYSICAL, "readFileObjectAt({0})", ifa);
        try {
            return baseImage.readFileObjectAt(ifa);
        } catch (Exception e) {
            throw new FdbDatabaseException("could not read object at " + ifa, e);
        }
    }

    public void updateFileObject(FdbObject fo) {
        tracer.println(Trace.PHYSICAL, "updateFileObject({0})", fo);
        try {
            baseImage.updateFileObject(fo.getIfa(), fo);
        } catch (Exception e) {
            throw new FdbDatabaseException("could not update object " + fo, e);
        }
    }

    public void deleteFileObject(InFileAddress ifa) {
        tracer.println(Trace.PHYSICAL, "deleteFileObject({0})", ifa);
        try {
            baseImage.deleteFileObject(ifa);
        } catch (Exception e) {
            throw new FdbDatabaseException("could not delete object at " + ifa, e);
        }
    }

    public void updateObjectInBase(InFileAddress ifa, Object o, boolean isUpdate) {
        FdbObject objectInBase = new FdbObjectImpl();
        objectInBase.setIfa(ifa);
        objectInBase.setData(o);
        if (isUpdate) {
            updateFileObject(objectInBase);
        } else {
            insertFileObject(objectInBase);
        }
    }


    public void start() {
        tracer.println(Trace.PHYSICAL, "start()");
        baseImage.start();
    }

    public void commit() {
        tracer.println(Trace.PHYSICAL, "commit()");
        baseImage.commit();
    }

    public void rollback() {
        tracer.println(Trace.PHYSICAL, "rollback()");
        baseImage.rollback();
    }


    public void internalInsertFileObject(InFileAddress ifa, FdbObject fo) {
        tracer.println(Trace.PHYSICAL, "internalInsertFileObject({0}, {1})", ifa, fo);
        int i = 0;
        try {
            RecordWriter rw = new RecordWriter(ifa.getAddress());
            rw.writeObject(fo);
            if (files.length > 1) {
                i = ifa.getFileIndex();
            }
            files[i].insertRecord(rw);
        } catch (Exception e) {
            Object v = null;
            try {
                v = files[i].readRecord(ifa.getAddress()).readObject();
            } catch (Exception e2) {
                throw new FdbDatabaseException("could not insert object " + fo, e);
            }
            throw new FdbDatabaseException("could not insert object " + fo +
                                           ";\n\t\tvalue already in base : " + String.valueOf(v), e);
        }
    }

    public FdbObject internalReadFileObjectAt(InFileAddress ifa) {
        tracer.println(Trace.PHYSICAL, "internalReadFileObjectAt({0})", ifa);
        FdbObject result = null;
        int i = 0;
        try {
            if (files.length > 1) {
                i = ifa.getFileIndex();
            }
            result = (FdbObject) files[i].readRecord(ifa.getAddress()).readObject();
            result.setIfa(ifa);
        } catch (Exception e) {
            throw new FdbDatabaseException("could not read object at " + ifa, e);
        }
        return result;
    }

    public void internalUpdateFileObject(InFileAddress ifa, FdbObject fo) {
        tracer.println(Trace.PHYSICAL, "internalUpdateFileObject({0}, {1})", ifa, fo);
        try {
            RecordWriter rw = new RecordWriter(ifa.getAddress());
            rw.writeObject(fo);
            int i = 0;
            if (files.length > 1) {
                i = ifa.getFileIndex();
            }
            files[i].updateRecord(rw);
        } catch (Exception e) {
            throw new FdbDatabaseException("could not update object " + fo, e);
        }
    }

    public void internalDeleteFileObject(InFileAddress ifa) {
        tracer.println(Trace.PHYSICAL, "internalDeleteFileObject({0})", ifa);
        try {
            int i = 0;
            if (files.length > 1) {
                i = ifa.getFileIndex();
            }
            files[i].deleteRecord(ifa.getAddress());
        } catch (Exception e) {
            throw new FdbDatabaseException("could not delete object at " + ifa, e);
        }
    }


    public Iterator getRAFFiles() {
        tracer.println(Trace.PHYSICAL, "getRAFFiles()");
        Collection result = new ArrayList();
        for (int i = 0; i < files.length; i++) {
            result.add(files[i]);
        }
        return result.iterator();
    }

    public InTransactionBaseImage getBaseImage() {
        return baseImage;
    }

    public String toString() {
        RecordsFileInspector rfi = new RecordsFileInspector();
        StringWriter sw = new StringWriter();
        PrintWriter out = new PrintWriter(sw);
        try {
            rfi.inspectRecordsFile(out, files[0]);
        } catch (Exception e) {
            out.println("error during inspection of the database... : " + e.getMessage());
        }
        out.flush();
        return sw.toString();
    }



    private InTransactionBaseImage baseImage;
    private RecordsFile[] files;

    private boolean isNewBase;

    private Trace tracer;


}
