package org.objectweb.fdb.impl.basic;

import org.objectweb.fdb.impl.trace.Trace;

import java.io.*;
import java.util.Enumeration;

public abstract class BaseRecordsFile {
    // Total length in bytes of the global database headers.
    protected static final int FILE_HEADERS_REGION_LENGTH = 16;

    // Number of bytes in the record header.
    protected static final int RECORD_HEADER_LENGTH = 16;

    // File pointer to the num records header.
    protected static final long NUM_RECORDS_HEADER_LOCATION = 0;

    // File pointer to the data start pointer header.
    protected static final long DATA_START_HEADER_LOCATION = 4;

    protected double additionalSpace;

    // The total length of one index entry - the key length plus the record header length.
    protected int indexEntryLength;

    // The length of a key in the index. default = 64
    protected int maxKeyLength;

    // Current file pointer to the start of the record data.
    protected long dataStartPtr;

    protected Trace tracer;

    // The underlying file.
    private RandomAccessFile file;

    /**
     * Creates a new database file, initializing the appropriate headers. Enough space is allocated in
     * the index for the specified initial size.
     */
    protected BaseRecordsFile(String dbPath, int initialSize, int maxKeyLength, int additionalSpace)
            throws IOException, RecordsFileException {
        File f = new File(dbPath);
        if (f.exists()) {
            throw new RecordsFileException("Database already exits: " + dbPath);
        }

        file = new RandomAccessFile(f, "rw");
        this.maxKeyLength = maxKeyLength;
        this.additionalSpace = (((double) additionalSpace) / 100) + 1;
        indexEntryLength = maxKeyLength + RECORD_HEADER_LENGTH;

        //		System.out.println(this.maxKeyLength + " " + indexEntryLength);
        dataStartPtr = indexPositionToKeyFp(initialSize); // Record Data Region starts were the
        setFileLength(dataStartPtr); // (i+1)th index entry would start.
        writeNumRecordsHeader(0);
        writeDataStartPtrHeader(dataStartPtr);
    }

    /**
     * Opens an existing database file and initializes the dataStartPtr. The accessFlags
     * parameter can be "r" or "rw" -- as defined in RandomAccessFile.
     */
    protected BaseRecordsFile(String dbPath, String accessFlags, int maxKeyLength, int additionalSpace)
            throws IOException, RecordsFileException {
        File f = new File(dbPath);
        if (!f.exists()) {
            throw new RecordsFileException("Database not found: " + dbPath);
        }

        file = new RandomAccessFile(f, accessFlags);
        this.maxKeyLength = maxKeyLength;
        this.additionalSpace = (((double) additionalSpace) / 100) + 1;
        indexEntryLength = maxKeyLength + RECORD_HEADER_LENGTH;
        dataStartPtr = readDataStartHeader();
    }

    public abstract int getNumRecords();

    public abstract Enumeration enumerateKeys();

    public abstract boolean recordExists(String key);

    public synchronized void close() throws IOException, RecordsFileException {
        try {
            file.close();
        } finally {
            file = null;
        }
    }

    public synchronized void deleteRecord(String key) throws RecordsFileException, IOException {
        RecordHeader delRec = keyToRecordHeader(key);
        int currentNumRecords = getNumRecords();

        if (getFileLength() == (delRec.getDataPointer() +
                delRec.getDataCapacity())) {
            // shrink file since this is the last record in the file
            setFileLength(delRec.getDataPointer());
        } else {
            RecordHeader previous = getRecordAt(delRec.getDataPointer() - 1);

            if (previous != null) {
                // append space of deleted record onto previous record
                previous.setDataCapacity(previous.getDataCapacity() + delRec.getDataCapacity());
                writeRecordHeaderToIndex(previous);
            } else {
                // target record is first in the file and is deleted by adding its space to
                // the second record.
                RecordHeader secondRecord = getRecordAt(delRec.getDataPointer() + (long) delRec.getDataCapacity());
                byte[] data = readRecordData(secondRecord);
                secondRecord.setDataPointer(delRec.getDataPointer());
                secondRecord.setDataCapacity(secondRecord.getDataCapacity() + delRec.getDataCapacity());
                writeRecordData(secondRecord, data);
                writeRecordHeaderToIndex(secondRecord);
            }
        }
        deleteEntryFromIndex(key, delRec, currentNumRecords);
    }

    public synchronized void insertRecord(RecordWriter rw) throws RecordsFileException, IOException {
        String key = rw.getKey();
        if (recordExists(key)) {
            throw new RecordsFileException("Key exists: " + key);
        }
        insureIndexSpace(getNumRecords() + 1);
        int length = (new Double(rw.getDataLength() * additionalSpace)).intValue();
        RecordHeader newRecord = allocateRecord(key, length);
        writeRecordData(newRecord, rw);
        addEntryToIndex(key, newRecord, getNumRecords());
    }

    public synchronized RecordReader readRecord(String key) throws RecordsFileException, IOException {
        byte[] data = readRecordData(key);
        return new RecordReader(key, data);
    }

    /**
     * Updates an existing record. If the new contents do not fit in the original record,
     * then the update is handled by deleting the old record and adding the new.
     */
    public synchronized void updateRecord(RecordWriter rw)
            throws RecordsFileException, IOException {
        RecordHeader header = keyToRecordHeader(rw.getKey());

        if (rw.getDataLength() > header.getDataCapacity()) {
            deleteRecord(rw.getKey());
            insertRecord(rw);
        } else {
            writeRecordData(header, rw);
            writeRecordHeaderToIndex(header);
        }
    }

    /**
     * Returns the record to which the target file pointer belongs - meaning the specified location
     * in the file is part of the record data of the RecordHeader which is returned.  Returns null if
     * the location is not part of a record. (O(n) mem accesses)
     */
    protected abstract RecordHeader getRecordAt(long targetFp) throws RecordsFileException;

    /**
     * Locates space for a new record of dataLength size and initializes a RecordHeader.
     */
    protected abstract RecordHeader allocateRecord(String key, int dataLength)
            throws RecordsFileException, IOException;

    protected abstract RecordHeader keyToRecordHeader(String key) throws RecordsFileException;

    protected void setFileLength(long l) throws IOException {
        file.setLength(l);
    }

    protected long getFileLength() throws IOException {
        return file.length();
    }

    /**
     * Appends an entry to end of index. Assumes that insureIndexSpace() has already been called.
     */
    protected void addEntryToIndex(String key, RecordHeader newRecord, int currentNumRecords)
            throws IOException, RecordsFileException {
        DbByteArrayOutputStream temp = new DbByteArrayOutputStream(maxKeyLength);
        (new DataOutputStream(temp)).writeUTF(key);

        if (temp.size() > maxKeyLength) {
            throw new RecordsFileException("Key " + key + " of size " + temp.size() +
                                           " is larger than permitted size of " +
                                           maxKeyLength + " bytes");
        }

        file.seek(indexPositionToKeyFp(currentNumRecords));
        temp.writeTo(file);
        file.seek(indexPositionToRecordHeaderFp(currentNumRecords));
        newRecord.write(file);
        newRecord.setIndexPosition(currentNumRecords);
        writeNumRecordsHeader(currentNumRecords + 1);
    }

    /**
     * Removes the record from the index. Replaces the target with the entry at the
     * end of the index.
     */
    protected void deleteEntryFromIndex(String key, RecordHeader header, int currentNumRecords)
            throws IOException, RecordsFileException {
        if (header.getIndexPosition() != (currentNumRecords - 1)) {
            String lastKey = readKeyFromIndex(currentNumRecords - 1);
            RecordHeader last = keyToRecordHeader(lastKey);
            last.setIndexPosition(header.getIndexPosition());
            file.seek(indexPositionToKeyFp(last.getIndexPosition()));
            file.writeUTF(lastKey);
            file.seek(indexPositionToRecordHeaderFp(last.getIndexPosition()));
            last.write(file);
        }
        writeNumRecordsHeader(currentNumRecords - 1);
    }

    /**
     * Returns a file pointer in the index pointing to the first byte
     * in the key located at the given index position.
     */
    protected long indexPositionToKeyFp(int pos) {
        return FILE_HEADERS_REGION_LENGTH + (indexEntryLength * pos);
    }

    // Checks to see if there is space for and additional index entry. If
    // not, space is created by moving records to the end of the file.
    protected void insureIndexSpace(int requiredNumRecords)
            throws RecordsFileException, IOException {
        int originalFirstDataCapacity;
        int currentNumRecords = getNumRecords();
        long endIndexPtr = indexPositionToKeyFp(requiredNumRecords);

        if ((endIndexPtr > getFileLength()) && (currentNumRecords == 0)) {
            setFileLength(endIndexPtr);
            dataStartPtr = endIndexPtr;
            writeDataStartPtrHeader(dataStartPtr);

            return;
        }

        while (endIndexPtr > dataStartPtr) {
            RecordHeader first = getRecordAt(dataStartPtr);
            byte[] data = readRecordData(first);
            first.setDataPointer(getFileLength());

            // If first.dataCapacity is set to the actual data count BEFORE resetting dataStartPtr,
            // and there is free space in 'first', then dataStartPtr will not be reset to the start
            // of the second record.  Capture the capacity first and use it to perform the reset.
            originalFirstDataCapacity = first.getDataCapacity();
            first.setDataCapacity(data.length);
            setFileLength(first.getDataPointer() + data.length);
            writeRecordData(first, data);
            writeRecordHeaderToIndex(first);
            dataStartPtr += originalFirstDataCapacity;
            writeDataStartPtrHeader(dataStartPtr);
        }
    }

    protected long readDataStartHeader() throws IOException {
        file.seek(DATA_START_HEADER_LOCATION);
        return file.readLong();
    }

    protected int readNumRecordsHeader() throws IOException {
        file.seek(NUM_RECORDS_HEADER_LOCATION);
        return file.readInt();
    }

    protected byte[] readRecordData(String key) throws IOException, RecordsFileException {
        return readRecordData(keyToRecordHeader(key));
    }

    protected byte[] readRecordData(RecordHeader header) throws IOException {
        byte[] buf = new byte[header.getDataCount()];
        file.seek(header.getDataPointer());
        file.readFully(buf);
        return buf;
    }

    protected void writeDataStartPtrHeader(long dataStartPtr) throws IOException {
        file.seek(DATA_START_HEADER_LOCATION);
        file.writeLong(dataStartPtr);
    }

    protected void writeNumRecordsHeader(int numRecords) throws IOException {
        file.seek(NUM_RECORDS_HEADER_LOCATION);
        file.writeInt(numRecords);
    }

    /**
     * Updates the contents of the given record. A RecordsFileException is thrown if the new data does not
     * fit in the space allocated to the record. The header's data count is updated, but not
     * written to the file.
     */
    protected void writeRecordData(RecordHeader header, RecordWriter rw) throws IOException, RecordsFileException {
        if (rw.getDataLength() > header.getDataCapacity()) {
            throw new RecordsFileException("Record data does not fit");
        }
        header.setDataCount(rw.getDataLength());
        file.seek(header.getDataPointer());
        rw.writeTo(file);
    }

    /**
     * Updates the contents of the given record. A RecordsFileException is thrown if the new data does not
     * fit in the space allocated to the record. The header's data count is updated, but not
     * written to the file.
     */
    protected void writeRecordData(RecordHeader header, byte[] data)
            throws IOException, RecordsFileException {
        if (data.length > header.getDataCapacity()) {
            throw new RecordsFileException("Record data does not fit");
        }
        header.setDataCount(data.length);
        file.seek(header.getDataPointer());
        file.write(data, 0, data.length);
    }

    protected void writeRecordHeaderToIndex(RecordHeader header) throws IOException {
        file.seek(indexPositionToRecordHeaderFp(header.getIndexPosition()));
        header.write(file);
        updateHeaderInMemIndex(header);
    }

    /**
     * Returns a file pointer in the index pointing to the first byte
     * in the record pointer located at the given index position.
     */
    long indexPositionToRecordHeaderFp(int pos) {
        return indexPositionToKeyFp(pos) + maxKeyLength;
    }

    /**
     * Reads the ith key from the index.
     */
    String readKeyFromIndex(int position) throws IOException {
        file.seek(indexPositionToKeyFp(position));
        return file.readUTF();
    }

    /**
     * Reads the ith record header from the index.
     */
    RecordHeader readRecordHeaderFromIndex(int position) throws IOException {
        file.seek(indexPositionToRecordHeaderFp(position));
        return RecordHeader.readHeader(file);
    }

    protected void updateHeaderInMemIndex(RecordHeader header) throws IOException {
    }

    public synchronized void restructure() throws IOException, RecordsFileException {
        long length = file.length();
        long moveAddress = dataStartPtr;
        long nextTargetAddress = dataStartPtr;
        while (length > nextTargetAddress) {
            RecordHeader header = getRecordAt(nextTargetAddress);
            nextTargetAddress = header.getDataPointer() + header.getDataCapacity();
            moveAddress = moveDataBack(header, moveAddress);
        }
        setFileLength(moveAddress);
    }

    // return next address to move data
    private synchronized long moveDataBack(RecordHeader header, long moveAddress) throws IOException, RecordsFileException {
        if (header.getDataPointer() > moveAddress) {
            byte[] buf = readRecordData(header);
            long maxAddress = header.getDataPointer() + header.getDataCapacity();
            int size = (new Double(buf.length * additionalSpace)).intValue();

            if ((moveAddress + size) > maxAddress) {
                size = (int) (maxAddress - moveAddress);
            }

            RecordHeader newHeader = new RecordHeader(moveAddress, size);
            writeRecordData(newHeader, buf);
            newHeader.setIndexPosition(header.getIndexPosition());
            writeRecordHeaderToIndex(newHeader);
            return newHeader.getDataPointer() + newHeader.getDataCapacity();

            // no move, just resize
        } else if (header.getDataPointer() == moveAddress) {
            int size = (new Double(header.getDataCount() * additionalSpace)).intValue();
            if (size > header.getDataCapacity()) {
                size = header.getDataCapacity();
            }
            header.setDataCapacity(size);
            writeRecordHeaderToIndex(header);
            return (header.getDataPointer() + header.getDataCapacity());
        } else {
            throw new RecordsFileException("moveDataBack address error...");
        }
    }

    public void setTracer(Trace tracer) {
        this.tracer = tracer;
    }
}
