package org.objectweb.fdb.impl.basic;

import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;

public class RecordsFile extends BaseRecordsFile {
    /**
     * Hashtable which holds the in-memory index. For efficiency, the entire index
     * is cached in memory. The hashtable maps a key of type String to a RecordHeader.
     */
    protected Hashtable memIndex;

    /**
     * Creates a new database file.  The initialSize parameter determines the
     * amount of space which is allocated for the index.  The index can grow
     * dynamically, but the parameter is provide to increase
     * efficiency.
     */
    public RecordsFile(String dbPath, int initialSize, int maxKeyLength, int additionalSpace)
            throws IOException, RecordsFileException {
        super(dbPath, initialSize, maxKeyLength, additionalSpace);
        memIndex = new Hashtable(initialSize);
    }

    /**
     * Opens an existing database and initializes the in-memory index.
     */
    public RecordsFile(String dbPath, String accessFlags, int maxKeyLength, int additionalSpace)
            throws IOException, RecordsFileException {
        super(dbPath, accessFlags, maxKeyLength, additionalSpace);
        readHeadersInMemIndex();
    }

    public synchronized int getNumRecords() {
        return memIndex.size();
    }

    public synchronized void close() throws IOException, RecordsFileException {
        try {
            super.close();
        } finally {
            memIndex.clear();
            memIndex = null;
        }
    }

    public synchronized Enumeration enumerateKeys() {
        return memIndex.keys();
    }

    public synchronized boolean recordExists(String key) {
        return memIndex.containsKey(key);
    }

    /**
     * Returns the record to which the target file pointer belongs - meaning the specified location
     * in the file is part of the record data of the RecordHeader which is returned.  Returns null if
     * the location is not part of a record. (O(n) mem accesses)
     */
    protected RecordHeader getRecordAt(long targetFp) throws RecordsFileException {
        Enumeration e = memIndex.elements();
        while (e.hasMoreElements()) {
            RecordHeader next = (RecordHeader) e.nextElement();
            if ((targetFp >= next.getDataPointer()) &&
                    (targetFp < (next.getDataPointer() +
                    (long) next.getDataCapacity()))) {
                return next;
            }
        }
        return null;
    }

    /**
     * Adds the new record to the in-memory index and calls the super class add
     * the index entry to the file.
     */
    protected void addEntryToIndex(String key, RecordHeader newRecord, int currentNumRecords)
            throws IOException, RecordsFileException {
        super.addEntryToIndex(key, newRecord, currentNumRecords);
        memIndex.put(key, newRecord);
    }

    /**
     * This method searches the file for free space and then returns a RecordHeader
     * which uses the space. (O(n) memory accesses)
     */
    protected RecordHeader allocateRecord(String key, int dataLength)
            throws RecordsFileException, IOException {
        // search for empty space
        RecordHeader newRecord = null;
        Enumeration e = memIndex.elements();

        while (e.hasMoreElements()) {
            RecordHeader next = (RecordHeader) e.nextElement();

            if (dataLength <= next.getFreeSpace()) {
                newRecord = next.split();
                writeRecordHeaderToIndex(next);
                break;
            }
        }

        if (newRecord == null) {
            // append record to end of file - grows file to allocate space
            long fp = getFileLength();
            setFileLength(fp + dataLength);
            newRecord = new RecordHeader(fp, dataLength);
        }

//        memIndex.put(key, newRecord);
        return newRecord;
    }

    /**
     * Removes the record from the index. Replaces the target with the entry at the
     * end of the index.
     */
    protected void deleteEntryFromIndex(String key, RecordHeader header, int currentNumRecords)
            throws IOException, RecordsFileException {
        super.deleteEntryFromIndex(key, header, currentNumRecords);
        memIndex.remove(key);
    }

    /**
     * Maps a key to a record header by looking it up in the in-memory index.
     */
    protected RecordHeader keyToRecordHeader(String key) throws RecordsFileException {
        RecordHeader h = (RecordHeader) memIndex.get(key);
        if (h == null) {
            throw new RecordsFileException("Key not found: " + key);
        }
        return h;
    }

    protected void readHeadersInMemIndex() throws IOException {
//        System.out.println("readHeadersInMemIndex");
        int numRecords = readNumRecordsHeader();
        memIndex = new Hashtable(numRecords);

        for (int i = 0; i < numRecords; i++) {
            String key = readKeyFromIndex(i);
            RecordHeader header = readRecordHeaderFromIndex(i);
            header.setIndexPosition(i);
            memIndex.put(key, header);
        }
    }

    protected void updateHeaderInMemIndex(RecordHeader header) throws IOException {
        memIndex.put(readKeyFromIndex(header.getIndexPosition()), header);
    }
}
