package org.objectweb.fdb.impl.impltx;

import org.objectweb.fdb.FdbLoid;
import org.objectweb.fdb.api.*;
import org.objectweb.fdb.exception.FdbDatabaseException;
import org.objectweb.fdb.exception.FdbIdentityException;
import org.objectweb.fdb.exception.FdbTransactionException;
import org.objectweb.fdb.impl.FdbPoid;
import org.objectweb.fdb.impl.FdbProperties;
import org.objectweb.fdb.impl.FdbSessionImpl;
import org.objectweb.fdb.impl.InternalFdbFactory;
import org.objectweb.fdb.impl.lock.DeadLockToken;
import org.objectweb.fdb.impl.lock.LockManagerOptimistic;
import org.objectweb.fdb.impl.lock.LockManagerPessimistic;
import org.objectweb.fdb.impl.trace.Trace;

import java.util.Collection;

public class FdbSessionTxImpl extends FdbSessionImpl implements FdbSessionTx {
    private FdbLockManager lockManager;
    private FdbLockTable lockTable;
    protected boolean isInTransaction = false;
    private volatile boolean sessionIsSleeping = false;
    private int timeOut;

    public FdbSessionTxImpl(FdbProperties properties) {
        super(properties);
        lockTable = InternalFdbFactory.getLockTable(properties);
        String txMode = properties.getProperty(FdbProperties.TX_MODE_KEY);
        if (txMode.equalsIgnoreCase(FdbLockManager.OPTIMISTIC_MODE)) {
            setOptimistic();
        } else if (txMode.equalsIgnoreCase(FdbLockManager.PESSIMISTIC_MODE)) {
            setPessimistic();
        } else {
            throw new FdbDatabaseException(FdbProperties.TX_MODE_KEY + " not valide in database properties...");
        }
        try {
            timeOut = (new Integer(properties.getProperty(FdbProperties.TIME_OUT_KEY))).intValue();
        } catch (Exception e) {
            throw new FdbDatabaseException(FdbProperties.TIME_OUT_KEY + " not valide in database properties...");
        }
        new DeadLockToken(sessionId);
    }

    public FdbOidTable getOidTable() {
        return oidTable;
    }

    // return a floid collection
    public Collection getExtent(String fullCLassName) {
        trace.println(Trace.SESSION, "getExtent(String fullCLassName)");
        Object clid = repository.getClidFromClassName(fullCLassName);
        Collection floids = oidTable.getAllFloidFromClid(clid);
        lockManager.checkLocksForSession(floids, sessionId);
        return floids;
    }

    public void setOptimistic() {
        trace.println(Trace.SESSION, "setOptimistic");
        if (isInTransaction) {
            throw new FdbTransactionException("could not change transaction mode during a transaction execution...");
        }
        lockManager = new LockManagerOptimistic();
        lockManager.setLockTable(lockTable);
        lockManager.setSession(this);
    }

    public void setPessimistic() {
        trace.println(Trace.SESSION, "setPessimistic");
        if (isInTransaction) {
            throw new FdbTransactionException("could not change transaction mode during a transaction execution...");
        }
        lockManager = new LockManagerPessimistic();
        lockManager.setLockTable(lockTable);
        lockManager.setSession(this);
    }

    public FdbProperties getProp() {
        return props;
    }

    public void begin() {
        trace.println(Trace.SESSION, "begin");
        if (isInTransaction) {
            throw new FdbTransactionException("the session " + sessionId + " is already in a transaction...");
        }
        isInTransaction = true;
        cache.clear();
        lockManager.begin();
    }

    public void commit() {
        trace.println(Trace.SESSION, "commit");
        if (!isInTransaction) {
            throw new FdbTransactionException("the session " + sessionId + " is not in a transaction...");
        }
        lockManager.commit();
        isInTransaction = false;
    }

    // return the oid
    public Object createObject(Object oid, Object[] objectToCreate) {
        FdbDataObject datas = InternalFdbFactory.getNewDataObject(props, objectToCreate);
        FdbLoid floid = getFloidFromOid(oid);
        if (!oidTable.contains(floid)) {
            throw new FdbIdentityException("the given oid " + oid + " doesn't exist in base...");
        }
        try {
            lockManager.prepareCreateObject(floid, datas);
            cache.put(floid, datas);
            lockManager.finishCreateObject(floid, datas);
        } catch (RuntimeException re) {
            lockManager.finishCreateObject(floid, datas);
            throw re;
        }
        return floid;
    }

    public void createObjectCommit(FdbLoid floid) {
        trace.println(Trace.SESSION, "createObjectCommit");
        FdbPoid fpoid = oidTable.getFpoid(floid);
        FdbDataObject datas = (FdbDataObject) cache.get(floid);
        if (datas == null) {
            datas = lockManager.getLockedValue(floid, sessionId);
        }
        fileAccess.insertDatas(fpoid, datas);
    }

    public void createObjectRollback(FdbLoid floid) {
        trace.println(Trace.SESSION, "createObjectRollback");
        if (cache.containsKey(floid)) {
            cache.remove(floid);
        }
        oidTable.removeFloid(floid);
    }

    public void deleteObjectByOid(Object oid) {
        trace.println(Trace.SESSION, "deleteObjectByOid");
        FdbLoid floid = getFloidFromOid(oid);
        if (!oidTable.contains(floid)) {
            throw new FdbIdentityException("the given oid " + oid + " doesn't exist in base...");
        }
        try {
            lockManager.prepareDeleteObjectByOid(floid, oidTable.getUpdate(floid));
            if (cache.containsKey(floid)) {
                cache.remove(floid);
            }
            lockManager.finishDeleteObjectByOid(floid);
        } catch (RuntimeException re) {
            lockManager.finishDeleteObjectByOid(floid);
            throw re;
        }
    }

    public void deleteObjectCommit(FdbLoid floid) {
        trace.println(Trace.SESSION, "deleteObjectCommit");
        FdbPoid fpoid = oidTable.getFpoid(floid);
        if (fpoid != null) {
            fileAccess.deleteDatas(fpoid);
            oidTable.removeFloid(floid);
        }
    }

    public void deleteObjectRollback(FdbLoid floid) {
        // nothing to do here
    }

    public void goToBed() {
        trace.println(Trace.SESSION, "goToBed");
        sessionIsSleeping = true;
    }

    public void inspectLocks() {
        lockTable.publishLocks();
    }

    public Object makeNewFileOid(Class objectClass) {
        return makeNewFileOid(objectClass.getName());
    }

    public Object makeNewFileOid(String objectClassName) {
        trace.println(Trace.SESSION, "makeNewFileOid");
        FdbLoid floid = oidProvider.allocateNewFloid();
        Object theClid = repository.getClidFromClassName(objectClassName);
        FdbPoid fpoid = fileAccess.allocateNewFpoid(theClid);
        oidTable.insertFpoid(floid, fpoid);
        lockManager.finishMakeNewFileOid(floid);
        return floid.getClone();
    }

    public void makeNewFileOidCommit(FdbLoid floid) {
        // nothing to do
    }

    public void makeNewFileOidRollback(FdbLoid floid) {
        if (!oidTable.contains(floid)) {
            throw new FdbIdentityException("the given oid " + floid + " doesn't exist in base...");
        }
        FdbPoid fpoid = oidTable.getFpoid(floid);
        oidTable.removeFloid(floid);
        if (fpoid != null) {
            fileAccess.desallocateFpoid(fpoid);
        } else {
            throw new FdbIdentityException("the given oid doesn't exist in base...");
        }
    }

    public Object[] readObjectByOid(Object oid, boolean withCache) {
        return readObjectByOidInDataObject(oid, true).getDatas();
    }

    public FdbDataObject readObjectByOidInDataObject(Object oid, boolean withCache) {
        trace.println(Trace.SESSION, "readObjectByOidInDataObject");
        FdbLoid floid = getFloidFromOid(oid);
        FdbDataObject result;
        lockManager.prepareReadObjectByOid(floid);
        if (oidTable.contains(floid)) {
            if (withCache && cache.containsKey(floid)) {
                result = (FdbDataObject) cache.get(floid);
            } else {
                FdbDataObject datas = null;
                // delete not commited for this session
                if (lockManager.containsDeletedValue(floid, sessionId)) {
                    lockManager.finishReadObjectByOid(floid, null, (short) -1);
                    throw new FdbIdentityException("the given oid " + oid + " doesn't exist in base...");
                    // create or update not commited for this session
                } else if (lockManager.containsLockedValue(floid, sessionId)) {
                    datas = lockManager.getLockedValue(floid, sessionId);
                    // object not modified by current session, read it in base
                } else {
                    FdbPoid fpoid = oidTable.getFpoid(floid);
                    if (fpoid != null) {
                        datas = fileAccess.readDatas(fpoid);
                    } else {
                        lockManager.finishReadObjectByOid(floid, null, (short) -1);
                        throw new FdbIdentityException("the given oid " + oid + " doesn't exist in base...");
                    }
                    if (datas == null) {
                        lockManager.finishReadObjectByOid(floid, null, (short) -1);
                        throw new FdbIdentityException("the given oid " + oid + " doesn't exist in base...");
                    }
                }
                if (withCache) {
                    cache.put(floid, datas);
                }
                result = datas;
            }
        } else {
            lockManager.finishReadObjectByOid(floid, null, (short) -1);
            throw new FdbIdentityException("the given oid " + oid + " doesn't exist in base...");
        }
        lockManager.finishReadObjectByOid(floid, result, oidTable.getUpdate(floid));
        return result;
    }

    public void rollback() {
        trace.println(Trace.SESSION, "rollback");
        if (!isInTransaction) {
            throw new FdbTransactionException("the session " + sessionId + " is not in a transaction...");
        }
        lockManager.rollback();
        isInTransaction = false;
    }

    public void sleep(FdbLoid floid) {
        sleep(floid, "");
    }

    public void sleep(FdbLoid floid, String message) {
        trace.println(Trace.SESSION, "sleep");
        if (sessionIsSleeping) {
            int i = 0;
            while (sessionIsSleeping && (i < timeOut)) {
                Thread.yield();
                i++;
            }
            if (i == timeOut) {
                sessionIsSleeping = false;
                lockTable.getLock(floid).resolveTimeOut(sessionId);
            }
        }
    }

    // return oid
    public Object updateObjectByOid(Object oid, Object[] objectToUpdate) {
        trace.println(Trace.SESSION, "updateObjectByOid");
        FdbLoid floid = getFloidFromOid(oid);
        if (!oidTable.contains(floid)) {
            throw new FdbIdentityException("the given oid " + oid + " doesn't exist in base...");
        }
        FdbDataObject datas = lockManager.getLockedValue(floid, sessionId);
        short update = -1;
        if (datas == null) {
            datas = InternalFdbFactory.getNewDataObject(props, objectToUpdate);
            update = oidTable.getUpdate(floid);
        }
        datas.setDatas(objectToUpdate);
        try {
            lockManager.prepareUpdateObjectByOid(floid, datas);
            cache.put(floid, datas);
            lockManager.finishUpdateObjectByOid(floid, datas, update);
        } catch (RuntimeException re) {
            lockManager.finishUpdateObjectByOid(floid, datas, update);
            throw re;
        }
        return floid.getClone();
    }

    public void updateObjectCommit(FdbLoid floid) {
        trace.println(Trace.SESSION, "updateObjectCommit");
        boolean findInCache = true;
        FdbDataObject datas = (FdbDataObject) cache.get(floid);
        if (datas == null) {
            findInCache = false;
            datas = lockManager.getLockedValue(floid, sessionId);
        }
        FdbPoid fpoid = oidTable.getFpoid(floid);
        if (fpoid != null) {
            FdbPoid updatedFpoid = fileAccess.updateDatas(fpoid, datas);
            if (!updatedFpoid.equals(fpoid)) {
                oidTable.updatePoid(floid, updatedFpoid);
            }
        } else {
            throw new FdbIdentityException("the given oid " + floid + " doesn't exist in base...");
        }
        if (!findInCache) {
            cache.put(floid, datas);
        }
    }

    public void updateObjectRollback(FdbLoid floid) {
        trace.println(Trace.SESSION, "updateObjectRollback");
        if (cache.containsKey(floid)) {
            cache.remove(floid);
        }
    }

    public void wakeUp() {
        trace.println(Trace.SESSION, "wakeUp");
        sessionIsSleeping = false;
    }
}
