package org.objectweb.fdb.impl.lock;

import org.objectweb.fdb.api.FdbDataObject;
import org.objectweb.fdb.api.FdbLockTable;
import org.objectweb.fdb.api.FdbSessionTx;
import org.objectweb.fdb.exception.FdbDeadLockException;
import org.objectweb.fdb.exception.FdbLockException;
import org.objectweb.fdb.exception.FdbSleepingException;
import org.objectweb.fdb.impl.InternalFdbFactory;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.NoSuchElementException;

public class Lock implements Serializable {
    private static final boolean WITH_HISTORIC = false;
    public static final int NO_LOCK = 0;
    public static final int READ_LOCK = 1;
    public static final int WRITE_LOCK = 2;
    public static final int RW_LOCK = 3;
    private HashMap currentStates;
    private LinkedList readSessions;
    private LinkedList sleepingSessions;

    private Object currentWorkingSession; // utiliser un Integer  -1 pour signifier null
    private Object writeSession;
    private int currentLock;

    public Lock() {
        currentLock = NO_LOCK;
        currentWorkingSession = new Integer(-1);
        currentStates = new HashMap();
        readSessions = new LinkedList();
        sleepingSessions = new LinkedList();
    }

    public synchronized int getCurrentAction(Object sessionId) {
        int result = InTransactionImage.NO_ACTION;

        if (currentStates.containsKey(sessionId)) {
            result = ((InTransactionImage) currentStates.get(sessionId)).getCurrentAction();
        }

        return result;
    }

    public synchronized FdbDataObject getModifiedValue(Object sessionId) {
        InTransactionImage image = (InTransactionImage) currentStates.get(sessionId);

        if (image != null) {
            return image.getCurrentValue();
        }

        return null;
    }

    public String getReadSessionsAsString() {
        return readSessions.toString();
    }

    public synchronized short getUpdate(Object sessionId) {
        InTransactionImage image = (InTransactionImage) currentStates.get(sessionId);

        if (image != null) {
            return image.getUpdate();
        }

        return (short) -1;
    }

    public String getWriteSessionAsString() {
        return writeSession.toString();
    }

    public synchronized void addModifiedValue(Object sessionId,
                                              FdbDataObject value, int action,
                                              short update) {
        // a read value can not be put over a write value
        InTransactionImage image = (InTransactionImage) currentStates.get(sessionId);

        if (image == null) {
            image = new InTransactionImage(action, value, update);
            currentStates.put(sessionId, image);
        } else {
            image.setValue(action, value);
            image.setUpdate(update);
        }

        //		if ((action != READ_ACTION) || (!hasModifiedValueNotReadWith(sessionId))) {
        //			// need to clone ???
        //			modifiedValues.put(sessionId, value);
        //			currentActions.put(sessionId, new Integer(action));
        //			table.saveLock(this);
        //		}
    }

    public synchronized void addSleepingSession(Object sessionId) {
        if (sleepingSessions.contains(sessionId)) {
            throw new FdbDeadLockException("sessionId = " + String.valueOf(sessionId));
        }
        DeadLockToken asleepingSession = DeadLockToken.getToken(sessionId);
        if (sleepingSessions.isEmpty()) {
            if (currentLock == READ_LOCK) {
                for (int i = 0; i < readSessions.size(); i++) {
                    asleepingSession.waitFor(DeadLockToken.getToken(readSessions.get(i)));
                }
            } else {
                asleepingSession.waitFor(DeadLockToken.getToken(currentWorkingSession));
            }
        } else {
            asleepingSession.waitFor(DeadLockToken.getToken(sleepingSessions.getLast()));
        }
        sleepingSessions.addLast(sessionId);
    }

    public synchronized boolean canBeDeleted() {
        return ((currentLock == NO_LOCK) && currentStates.isEmpty() &&
                sleepingSessions.isEmpty());

        //		return ((currentLock == NO_LOCK) && currentStates.isEmpty());
    }

    public synchronized boolean hasModifiedValueWith(Object sessionId) {
        return (currentStates.get(sessionId) != null);
    }

    public synchronized boolean hasReadLockWith(Object sessionId) {
        return readSessions.contains(sessionId);
    }

    public synchronized boolean hasToBeRemoveForExtentWith(Object sessionId) {
        if ((currentLock == WRITE_LOCK) && (!writeSession.equals(sessionId))) {
            return true;
        }

        if (currentStates.containsKey(sessionId)) {
            InTransactionImage itm = (InTransactionImage) currentStates.get(sessionId);

            return (itm.getCurrentAction() == InTransactionImage.NEW_OID_ACTION);
        } else {
            if (currentStates.size() > 0) {
                Iterator itms = currentStates.values().iterator();

                while (itms.hasNext()) {
                    InTransactionImage itm = (InTransactionImage) itms.next();

                    if (itm.getUpdate() == -1) {
                        return true;
                    }
                }

                return false;
            }

            return true;
        }
    }

    /********************* TESTS ********************************/
    public synchronized boolean hasWriteLockWith(Object sessionId) {
        return ((writeSession != null) && (writeSession.equals(sessionId)));
    }

    public synchronized void releaseAllLock(Object sessionId) {
        if (currentWorkingSession.equals(sessionId)) {
            if (currentLock == WRITE_LOCK) {
                if (WITH_HISTORIC) {
                    addToHistory("releaseAllLock : " + sessionId +
                                 " release WL)");
                }

                writeSession = null;
            } else if (currentLock == READ_LOCK) {
                readSessions.remove(sessionId);

                if (!readSessions.isEmpty()) {
                    if (WITH_HISTORIC) {
                        addToHistory("releaseAllLock : " + sessionId +
                                     " release a RL");
                    }

                    currentStates.remove(sessionId);

                    //					wakeUpASession(sessionId);
                    currentWorkingSession = readSessions.getFirst();

                    if (WITH_HISTORIC) {
                        addToHistory("releaseAllLock : " + sessionId +
                                     " release a RL, current become " +
                                     currentWorkingSession);
                    }

                    wakeUpASession(sessionId);

                    return;
                }
            } else if (currentLock == RW_LOCK) {
                if (WITH_HISTORIC) {
                    addToHistory("releaseAllLock : " + sessionId +
                                 " release a RWL");
                }

                readSessions.remove(sessionId);
                writeSession = null;
            } else if (currentLock == NO_LOCK) {
                // nothing to do here
            } else {
                throw new FdbLockException();
            }

            currentStates.remove(sessionId);
            currentLock = NO_LOCK;
            currentWorkingSession = new Integer(-1);
            wakeUpASession(sessionId);
        } else {
            // currentWorkingSession != sessionId
            if (readSessions.contains(sessionId)) {
                //				wakeUpASession(sessionId);
                readSessions.remove(sessionId);
                wakeUpASession(sessionId);
            }

            currentStates.remove(sessionId);

            if (WITH_HISTORIC) {
                addToHistory("releaseAllLock : " + sessionId +
                             " release but is not currentWorkingSession");
            }
        }
    }

    public synchronized void releaseReadLock(Object sessionId) {
        if (currentLock == READ_LOCK) {
            readSessions.remove(sessionId);

            //			wakeUpASession(sessionId);
            if (readSessions.isEmpty()) {
                currentLock = NO_LOCK;
                currentWorkingSession = new Integer(-1);

                if (WITH_HISTORIC) {
                    addToHistory("releaseReadLock : " + sessionId +
                                 " release unique RL");
                }
            } else {
                if (currentWorkingSession.equals(sessionId)) {
                    currentWorkingSession = readSessions.getFirst();
                }

                if (WITH_HISTORIC) {
                    addToHistory("releaseReadLock : " + sessionId +
                                 " release a RL, still : " + readSessions +
                                 " CWS : " + currentWorkingSession);
                }
            }

            wakeUpASession(sessionId);
        } else if (currentLock == RW_LOCK) {
            readSessions.remove(sessionId);
            currentLock = WRITE_LOCK;

            if (WITH_HISTORIC) {
                addToHistory("releaseReadLock : " + sessionId +
                             " release RWL to WL)");
            }
        } else {
            System.out.println("EXCEPTION during releaseReadLock with " +
                               sessionId + " : " + this);
            throw new FdbLockException();
        }
    }

    public synchronized void resolveTimeOut(Object sessionId) {
        //		System.out.println("time out " + sessionId);
        //		System.out.println("BEFORE "+DeadLockToken.getTokens());
        //		String debug = "BEFORE " + DeadLockToken.getTokens().toString();
        DeadLockToken timeOutToken = DeadLockToken.getToken(sessionId);

        if (sleepingSessions.contains(sessionId)) {
            int index = sleepingSessions.indexOf(sessionId);

            if ((index + 1) < sleepingSessions.size()) {
                DeadLockToken tokenWhoWaitForTimeOutToken = DeadLockToken.getToken(sleepingSessions.get(index +
                                                                                                        1));
                timeOutToken.timeOutWakeUp(tokenWhoWaitForTimeOutToken);
            } else {
                timeOutToken.timeOutWakeUp(null);
            }

            sleepingSessions.remove(sessionId);
        } else {
            // dans quels cas ?????
            DeadLockToken tokenWhoWaitForTimeOutToken = null;

            try {
                tokenWhoWaitForTimeOutToken = DeadLockToken.getToken(sleepingSessions.getFirst());
            } catch (NoSuchElementException e) {
                //				System.out.println(sessionId);
                //				System.out.println(this);
                //				System.out.println(debug);
                //				System.out.println("AFTER " + DeadLockToken.getTokens());
            }

            timeOutToken.timeOutWakeUp(tokenWhoWaitForTimeOutToken);
        }

        //		System.out.println("AFTER "+DeadLockToken.getTokens());
        //		System.out.println("");
    }

    /*************************** TAKE RELEASE ADD REMOVE GET SET ****************************************/
    public synchronized void takeReadLock(Object sessionId) {
        // 1 : cas pas de lock et pas de sleeping -> ajouter RL avec sessionId
        // 2 : cas pas de lock et sleepings et le premier sleeping attends la session courante -> add RL
        // 3 : cas pas de lock et sleepings et le premier sleeping n'attends pas la session courante -> sleep
        if (currentLock == NO_LOCK) {
            if (sleepingSessions.isEmpty()) { // 1
                currentLock = READ_LOCK;

                if (WITH_HISTORIC) {
                    addToHistory("[case 1] " + sessionId +
                                 " take read lock on (no lock, no sleeping)");
                }
            } else {
                DeadLockToken awakingToken = DeadLockToken.getToken(sleepingSessions.getFirst());

                if (awakingToken.isWaitingFor(DeadLockToken.getToken(sessionId))) { // 2
                    currentLock = READ_LOCK;

                    if (WITH_HISTORIC) {
                        addToHistory("[case 2] " + sessionId +
                                     " take read lock on (no lock, sleeping waiting for this session)");
                    }
                } else { // 3

                    if (WITH_HISTORIC) {
                        addToHistory("[case 3] " + sessionId +
                                     " sleep for read lock on (no lock, sleepings " +
                                     sleepingSessions + ")");
                    }

                    throw new FdbSleepingException(this);
                }
            }

            // 4 : cas read session avec sessionId et eventuellement d'autre -> ne fe rien
            // 5 : cas read session avec autre exclusivement que sessionId et pas sleeping -> ajouter sessionId readSessions
            // 6 : cas read session avec autre exclusivement que sessionId et sleeping -> sleep
        } else if (currentLock == READ_LOCK) {
            if ((!readSessions.contains(sessionId)) &&
                    (!sleepingSessions.isEmpty())) { // 6

                if (WITH_HISTORIC) {
                    addToHistory("[case 6] " + sessionId +
                                 " sleep for read lock on (other read locks, sleepings)");
                }

                throw new FdbSleepingException(this);
            } else { // 4 & 5

                if (WITH_HISTORIC) {
                    addToHistory("[case 4&5] " + sessionId +
                                 " take read lock on existing RL. Change CWS to sessionId");
                }
            }

            // 7 : cas write session avec sessionId	-> transforme Wl en RWL
            // 8 : cas write session avec autre		-> sleep
        } else if (currentLock == WRITE_LOCK) {
            if (currentWorkingSession.equals(sessionId)) { // 7

                if (WITH_HISTORIC) {
                    addToHistory("[case 7] " + sessionId +
                                 " take update to RWL");
                }

                currentLock = RW_LOCK;
            } else { // 8

                if (WITH_HISTORIC) {
                    addToHistory("[case 8] " + sessionId +
                                 " sleep for read lock on (other WL : " +
                                 currentWorkingSession + ")");
                }

                throw new FdbSleepingException(this);
            }

            // 9 : cas RW session et working = sessionId -> rien  faire
            // 10 : cas RW session et working != sessionId -> sleep
        } else if (currentLock == RW_LOCK) {
            if (currentWorkingSession.equals(sessionId)) { // 9

                if (WITH_HISTORIC) {
                    addToHistory("[case 9] " + sessionId +
                                 " sleep for read lock on (other RWL)");
                }
            } else { // 10

                if (WITH_HISTORIC) {
                    addToHistory("[case 10] " + sessionId +
                                 " sleep for read lock on (other RWL)");
                }

                throw new FdbSleepingException(this);
            }
        }

        if (!readSessions.contains(sessionId)) {
            readSessions.addLast(sessionId);
        }

        currentWorkingSession = sessionId;
    }

    public synchronized void takeWriteLock(Object sessionId) {
        // 1 : cas pas de lock et pas sleepings -> ajouter WL avec sessionId
        // 2 : cas pas de lock et sleepings et premier sleeping attends sessionId -> sessionId pose WL
        // 3 : cas pas de lock et sleepings et premier sleeping n'attends pas sessionId -> sleep
        if (currentLock == NO_LOCK) {
            if (sleepingSessions.isEmpty()) { // 1

                if (WITH_HISTORIC) {
                    addToHistory("[case 1] " + sessionId +
                                 " take write lock on (no lock, no sleeping)");
                }

                currentLock = WRITE_LOCK;
            } else {
                DeadLockToken awakingToken = DeadLockToken.getToken(sleepingSessions.getFirst());

                if (awakingToken.isWaitingFor(DeadLockToken.getToken(sessionId))) { // 2
                    currentLock = WRITE_LOCK;

                    if (WITH_HISTORIC) {
                        addToHistory("[case 2] " + sessionId +
                                     " take write lock on (no lock, sleeping waiting for this session)");
                    }
                } else { // 3

                    if (WITH_HISTORIC) {
                        addToHistory("[case 3] " + sessionId +
                                     " sleep for write lock on (no lock, sleepings " +
                                     sleepingSessions + ")");
                    }

                    throw new FdbSleepingException(this);
                }
            }

            // 4 : cas read session avec seulement sessionId -> transforme RL en RWL
            // 5 : cas read session avec autre exclusivement que sessionId -> sleep
            // 6 : cas read session avec autre que sessionId ET sessionId -> retirer session des reads et sleep
        } else if (currentLock == READ_LOCK) {
            if (readSessions.contains(sessionId)) {
                if (readSessions.size() > 1) { // 6
                    readSessions.remove(sessionId);
                    wakeUpASession(sessionId); // ?????????????????????????????
                    throw new FdbSleepingException(this);
                } else { // 4
                    currentLock = RW_LOCK;

                    if (WITH_HISTORIC) {
                        addToHistory("[case 4] " + sessionId +
                                     " update RL to RWL)");
                    }
                }
            } else { // 5

                if (WITH_HISTORIC) {
                    addToHistory("[case 5] " + sessionId +
                                 " sleep for write lock on (exclusively other read session : " +
                                 readSessions + ")");
                }

                throw new FdbSleepingException(this);
            }

            // 7 : cas RL/RWL session avec sessionId	-> ne fait rien
            // 8 : cas RL/RWL session avec autre		-> sleep
        } else { // currentLock == RL ou RWL

            if (currentWorkingSession.equals(sessionId)) { // 7

                if (WITH_HISTORIC) {
                    addToHistory("[case 7] " + sessionId +
                                 " take write lock but have already it)");
                }

                return;
            } else { // 8

                if (WITH_HISTORIC) {
                    addToHistory("[case 8] " + sessionId +
                                 " sleep for write lock on (other WL : " +
                                 writeSession + ")");
                }

                throw new FdbSleepingException(this);
            }
        }

        writeSession = sessionId;
        currentWorkingSession = sessionId;
    }

    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append(getLockAsString(currentLock)).append(" ")
                .append(super.toString()).append("\n");
        sb.append("\tRead Locks : ").append(readSessions).append("\n");
        sb.append("\tWrite Lock : ").append(writeSession).append("\n");
        sb.append("\tWorking Lock : ").append(currentWorkingSession).append("\n");
        sb.append("\tActions : \n");
        sb.append("\tSleeping session : ").append(sleepingSessions).append("\n");

        Iterator sessions = currentStates.keySet().iterator();

        if (!sessions.hasNext()) {
            sb.append("CurrentStates is empty");
        }

        while (sessions.hasNext()) {
            Object key = sessions.next();
            InTransactionImage image = (InTransactionImage) currentStates.get(key);
            sb.append(" ").append(image);
        }

        sb.append("\n");

        //		sb.append("HISTORIC : ").append(historic).append("\n");
        return sb.toString();
    }

    public void update(Object sessionId) {
        ((InTransactionImage) currentStates.get(sessionId)).update();
    }

    private String getLockAsString(int lock) {
        if (lock == NO_LOCK) {
            return "no_lock";
        } else if (lock == READ_LOCK) {
            return "r_lock";
        } else if (lock == WRITE_LOCK) {
            return "w_lock";
        } else {
            return "rw_lock";
        }
    }

    /********************************* UTIL ********************************************/
    private void addToHistory(Object message) {
        //		historic.add(message);
    }

    private synchronized void wakeUpASession(Object releasingSessionId) {
        if (!sleepingSessions.isEmpty()) {
            Object awakingId = sleepingSessions.getFirst();
            DeadLockToken releasingSessionToken = DeadLockToken.getToken(releasingSessionId);
            DeadLockToken awakingSessionToken = DeadLockToken.getToken(awakingId);

            synchronized (awakingSessionToken) {
                if (WITH_HISTORIC) {
                    addToHistory(releasingSessionId.toString() + " wake up " +
                                 awakingId);
                }

                awakingSessionToken.noMoreWaitFor(releasingSessionToken);

                if (awakingSessionToken.couldWakeUp()) {
                    sleepingSessions.remove(awakingId);
                    ((FdbSessionTx) InternalFdbFactory.getSessionById(awakingId)).wakeUp();
                }
            }
        }
    }
}
