package org.objectweb.fdb.impl.lock;

import org.objectweb.fdb.FdbLoid;
import org.objectweb.fdb.api.FdbDataObject;
import org.objectweb.fdb.exception.FdbSleepingException;
import org.objectweb.fdb.exception.FdbTransactionException;
import org.objectweb.fdb.exception.FdbDeadLockException;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

public class LockManagerOptimistic extends LockManagerDefault {
    private static boolean isCommitting = false;

    public void prepareReadObjectByOid(FdbLoid floid) {
        boolean isLockTaken = false;
        boolean isAsleeping = false;
        Lock lock;
        while (!isLockTaken) {
            synchronized (lockTable) {
                lock = lockTable.getLock(floid);
                synchronized (lock) {
                    try {
                        lock.takeReadLock(sessionId);
                        isLockTaken = true;
                    } catch (FdbSleepingException fse) {
                        session.goToBed();
                        lock.addSleepingSession(sessionId);
                        isAsleeping = true;
                    }
                }
            }
            if (isAsleeping) {
                session.sleep(floid, "during prepareReadObjectByOid");
                isAsleeping = false;
            }
        }
    }

    public void prepareDeleteObjectByOid(FdbLoid floid, short update) {
        lockTable.getLock(floid).addModifiedValue(sessionId, null,
                                                  InTransactionImage.DELETE_ACTION,
                                                  update);
    }

    public void finishCreateObject(FdbLoid floid, FdbDataObject objectToCreate) {
        lockTable.getLock(floid).addModifiedValue(sessionId, objectToCreate,
                                                  InTransactionImage.CREATE_ACTION,
                                                  (short) -1);
    }

    public void finishMakeNewFileOid(FdbLoid floid) {
        lockTable.getLock(floid).addModifiedValue(sessionId, null,
                                                  InTransactionImage.NEW_OID_ACTION,
                                                  (short) -1);
    }

    public void finishReadObjectByOid(FdbLoid floid, FdbDataObject objectToRead, short update) {
        Lock lock = lockTable.getLock(floid);
        lock.addModifiedValue(sessionId, objectToRead, InTransactionImage.READ_ACTION, update);
        lock.releaseReadLock(sessionId);
    }

    public void finishUpdateObjectByOid(FdbLoid floid, FdbDataObject objectToUpdate, short update) {
        lockTable.getLock(floid).addModifiedValue(sessionId, objectToUpdate,
                                                  InTransactionImage.UPDATE_ACTION,
                                                  update);
    }

    public synchronized void commit() {
        if (isCommitting) {
            waitForCommit();
        }
        isCommitting = true;
        boolean hasError = true;

        try {
            ArrayList usefullLocks = new ArrayList();
            ArrayList usefullFloid = new ArrayList();

            while (!getUsedLock(usefullFloid, usefullLocks)) {
                session.sleep((FdbLoid) usefullFloid.get(0), "during commit");
            }

            // we have all needed lock...
            Collection locksToRelease = new ArrayList();
            Collection floidsToRemove = new ArrayList();

            for (int i = 0; i < usefullLocks.size(); i++) {
                Lock lock = (Lock) usefullLocks.get(i);
                FdbLoid floid = (FdbLoid) usefullFloid.get(i);

                if (lock.getCurrentAction(sessionId) == InTransactionImage.UPDATE_ACTION) {
                    checkUpdate(floid, lock.getUpdate(sessionId));
                    lock.update(sessionId);
                    session.updateObjectCommit(floid);
                    session.getOidTable().setUpdate(floid, lock.getUpdate(sessionId));
                } else if (lock.getCurrentAction(sessionId) == InTransactionImage.CREATE_ACTION) {
                    checkUpdate(floid, lock.getUpdate(sessionId));
                    lock.update(sessionId);
                    session.createObjectCommit(floid);
                    session.getOidTable().setUpdate(floid, lock.getUpdate(sessionId));
                } else if (lock.getCurrentAction(sessionId) == InTransactionImage.DELETE_ACTION) {
                    checkUpdate(floid, lock.getUpdate(sessionId));
                    lock.update(sessionId);
                    session.deleteObjectCommit(floid);
                } else if (lock.getCurrentAction(sessionId) == InTransactionImage.NEW_OID_ACTION) {
                    session.makeNewFileOidCommit(floid);
                }

                locksToRelease.add(lock);
                floidsToRemove.add(floid);
            }

            Iterator it = locksToRelease.iterator();
            while (it.hasNext()) {
                ((Lock) it.next()).releaseAllLock(sessionId);
            }
            it = floidsToRemove.iterator();
            while (it.hasNext()) {
                lockTable.removeLockByFloid((FdbLoid) it.next());
            }

            session.getFileAccess().getPhysicalAccess().commit();
            hasError = false;
        } finally {
            //			if (hasError) {
            //				System.out.println("error");
            //			}
            isCommitting = false;
            //			Thread.yield();
        }
    }

    public synchronized void rollback() {
        if (isCommitting) {
            waitForCommit();
        }
        isCommitting = true;
        boolean hasError = true;

        try {
            ArrayList usefullLocks = new ArrayList();
            ArrayList usefullFloid = new ArrayList();

            while (!getUsedLock(usefullFloid, usefullLocks)) {
                session.sleep((FdbLoid) usefullFloid.get(0), "during rollback");
            }

            // we have all needed lock...
            Collection locksToRelease = new ArrayList();
            Collection floidsToRemove = new ArrayList();

            for (int i = 0; i < usefullLocks.size(); i++) {
                Lock lock = (Lock) usefullLocks.get(i);
                FdbLoid floid = (FdbLoid) usefullFloid.get(i);

                if (lock.getCurrentAction(sessionId) == InTransactionImage.CREATE_ACTION) {
                    session.createObjectRollback(floid);
                } else if (lock.getCurrentAction(sessionId) == InTransactionImage.UPDATE_ACTION) {
                    session.updateObjectRollback(floid);
                } else if (lock.getCurrentAction(sessionId) == InTransactionImage.DELETE_ACTION) {
                    session.deleteObjectRollback(floid);
                } else if (lock.getCurrentAction(sessionId) == InTransactionImage.NEW_OID_ACTION) {
                    session.makeNewFileOidRollback(floid);
                }

                locksToRelease.add(lock);
                floidsToRemove.add(floid);
            }

            Iterator it = locksToRelease.iterator();
            while (it.hasNext()) {
                ((Lock) it.next()).releaseAllLock(sessionId);
            }
            it = floidsToRemove.iterator();
            while (it.hasNext()) {
                lockTable.removeLockByFloid((FdbLoid) it.next());
            }

            session.getFileAccess().getPhysicalAccess().rollback();
            hasError = false;
        } finally {
            //			if (hasError) {
            //				System.out.println("error");
            //			}
            isCommitting = false;
            //			Thread.yield();
        }
    }

    /******************************************************************************************/

    // floidList, lockList must be empty
    private boolean getUsedLock(List floidList, List lockList) {
        floidList.clear();
        lockList.clear();
        boolean hasException = false;
        Iterator it = lockTable.getKeys();
        while (it.hasNext()) {
            FdbLoid floid = (FdbLoid) it.next();
            synchronized (lockTable) {
                Lock lock = lockTable.getLock(floid);
                synchronized (lock) {
                    try {
                        if (lock.hasModifiedValueWith(sessionId)) {
                            lock.takeWriteLock(sessionId);
                            lockList.add(lock);
                            floidList.add(floid);
                        }
                    } catch (FdbSleepingException fse) {
                        hasException = true;
                        session.goToBed();
                        try {
                            lock.addSleepingSession(sessionId);
                        } catch (FdbDeadLockException dle) {
                        } finally {
                            floidList.clear();
                            lockList.clear();
                            floidList.add(floid);
                            hasException = true;
                        }

                    }
                }
            }
        }
        return (!hasException);
    }

    private void checkUpdate(FdbLoid floid, short currentUpdate) {
        short update = session.getOidTable().getUpdate(floid);
        if ((update == -1) && (currentUpdate == 0)) {
            return;
        }
        if (currentUpdate != update) {
            throw new FdbTransactionException("object have been modified since you get it...");
        }
    }

    private void waitForCommit() {
        //		System.out.println("session " + session.getSessionId() + " is waiting for committing");
        while (isCommitting) {
            Thread.yield();
        }
        //		System.out.println("session " + session.getSessionId() + " has finish to wait");
    }
}
