package org.objectweb.fdb.impl.lock;

import org.objectweb.fdb.FdbLoid;
import org.objectweb.fdb.api.FdbDataObject;
import org.objectweb.fdb.exception.FdbSleepingException;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

public class LockManagerPessimistic extends LockManagerDefault {
    private static final int READ = 0;
    private static final int WRITE = 1;
    private static boolean isCommitting = false;

    public void commit() {
        if (isCommitting) {
            waitForCommit();
        }

        isCommitting = true;

        boolean hasError = true;

        try {
            Iterator it = lockTable.getKeys();
            Collection locksToRelease = new ArrayList();
            Collection floidsToRemove = new ArrayList();

            while (it.hasNext()) {
                FdbLoid floid = (FdbLoid) it.next();
                floidsToRemove.add(floid);

                Lock lock = lockTable.getLock(floid);

                if (lock.hasWriteLockWith(sessionId)) {
                    if (lock.getCurrentAction(sessionId) == InTransactionImage.CREATE_ACTION) {
                        session.createObjectCommit(floid);
                    } else if (lock.getCurrentAction(sessionId) == InTransactionImage.UPDATE_ACTION) {
                        lock.update(sessionId);
                        session.updateObjectCommit(floid);
                    } else if (lock.getCurrentAction(sessionId) == InTransactionImage.DELETE_ACTION) {
                        session.deleteObjectCommit(floid);
                    } else if (lock.getCurrentAction(sessionId) == InTransactionImage.NEW_OID_ACTION) {
                        session.makeNewFileOidCommit(floid);
                    }

                    locksToRelease.add(lock);
                } else if (lock.hasReadLockWith(sessionId)) {
                    locksToRelease.add(lock);
                }
            }

            it = locksToRelease.iterator();

            while (it.hasNext()) {
                ((Lock) it.next()).releaseAllLock(sessionId);
            }

            it = floidsToRemove.iterator();

            while (it.hasNext()) {
                lockTable.removeLockByFloid((FdbLoid) it.next());
            }

            session.getFileAccess().getPhysicalAccess().commit();
            hasError = false;
        } finally {
            //			if (hasError) {
            //				System.out.println("error");
            //			}
            isCommitting = false;
            Thread.yield();
        }
    }

    public void finishCreateObject(FdbLoid floid, FdbDataObject objectToCreate) {
        lockTable.getLock(floid).addModifiedValue(sessionId, objectToCreate,
                                                  InTransactionImage.CREATE_ACTION,
                                                  (short) 0);
    }

    public void finishReadObjectByOid(FdbLoid floid, FdbDataObject objectToRead,
                                      short update) {
        lockTable.getLock(floid).addModifiedValue(sessionId, objectToRead,
                                                  InTransactionImage.READ_ACTION,
                                                  update);
    }

    public void finishUpdateObjectByOid(FdbLoid floid,
                                        FdbDataObject objectToUpdate, short update) {
        //		objectToUpdate.update();
        lockTable.getLock(floid).addModifiedValue(sessionId, objectToUpdate,
                                                  InTransactionImage.UPDATE_ACTION,
                                                  update);
    }

    public void prepareCreateObject(FdbLoid floid, FdbDataObject objectToCreate) {
        addLock(floid, "finishCreateObject", WRITE);
    }

    public void prepareDeleteObjectByOid(FdbLoid floid, short update) {
        addLock(floid, "finishDeleteObjectByOid", WRITE);
        lockTable.getLock(floid).addModifiedValue(sessionId, null,
                                                  InTransactionImage.DELETE_ACTION,
                                                  (short) 0);
    }

    public void prepareReadObjectByOid(FdbLoid floid) {
        addLock(floid, "prepareReadObjectByOid", READ);
    }

    public void prepareUpdateObjectByOid(FdbLoid floid,
                                         FdbDataObject objectToUpdate) {
        addLock(floid, "finishUpdateObjectByOid", WRITE);
    }

    public void rollback() {
        if (isCommitting) {
            waitForCommit();
        }

        isCommitting = true;

        boolean hasError = true;

        try {
            Iterator it = lockTable.getKeys();
            Collection locksToRelease = new ArrayList();
            Collection floidsToRemove = new ArrayList();

            while (it.hasNext()) {
                FdbLoid floid = (FdbLoid) it.next();
                floidsToRemove.add(floid);

                Lock lock = lockTable.getLock(floid);

                if (lock.hasWriteLockWith(sessionId)) {
                    if (lock.getCurrentAction(sessionId) == InTransactionImage.CREATE_ACTION) {
                        session.createObjectRollback(floid);
                    } else if (lock.getCurrentAction(sessionId) == InTransactionImage.UPDATE_ACTION) {
                        session.updateObjectRollback(floid);
                    } else if (lock.getCurrentAction(sessionId) == InTransactionImage.DELETE_ACTION) {
                        session.deleteObjectRollback(floid);
                    } else if (lock.getCurrentAction(sessionId) == InTransactionImage.NEW_OID_ACTION) {
                        session.makeNewFileOidRollback(floid);
                    }

                    locksToRelease.add(lock);
                } else if (lock.hasReadLockWith(sessionId)) {
                    locksToRelease.add(lock);
                }
            }

            it = locksToRelease.iterator();

            while (it.hasNext()) {
                ((Lock) it.next()).releaseAllLock(sessionId);
            }

            it = floidsToRemove.iterator();

            while (it.hasNext()) {
                lockTable.removeLockByFloid((FdbLoid) it.next());
            }

            session.getFileAccess().getPhysicalAccess().rollback();
            hasError = false;
        } finally {
            //			if (hasError) {
            //				System.out.println("error");
            //			}
            isCommitting = false;
            Thread.yield();
        }
    }

    private void addLock(FdbLoid floid, String message, int mode) {
        boolean isLockTaken = false;
        boolean isAsleeping = false;

        while (!isLockTaken) {
            Lock lock = lockTable.getLock(floid);

            synchronized (lockTable) {
                synchronized (lock) {
                    try {
                        if (mode == READ) {
                            lock.takeReadLock(sessionId);
                        } else {
                            lock.takeWriteLock(sessionId);
                        }

                        isLockTaken = true;
                    } catch (FdbSleepingException fse) {
                        session.goToBed();
                        lock.addSleepingSession(sessionId);
                        isAsleeping = true;
                    }
                }
            }

            if (isAsleeping) {
                session.sleep(floid, message);
                isAsleeping = false;
            }
        }
    }

    private void waitForCommit() {
        //		System.out.println("session " + session.getSessionId() + " is waiting for committing");
        while (isCommitting) {
            Thread.yield();
        }

        //		System.out.println("session " + session.getSessionId() + " has finish to wait");
    }
}
