package org.objectweb.fdb.impl.meta;

import org.objectweb.fdb.impl.meta.comparator.FdbFielComparator;
import org.objectweb.fdb.impl.InFileAddress;
import org.objectweb.fdb.exception.FdbSchemaException;

import java.io.Serializable;

public class ClassMetaDescription implements Serializable {

    public ClassMetaDescription(String fullClassName) {
        this.className = fullClassName;
        this.fields = new FieldDescription[0];
    }

    public String getClassName() {
        return className;
    }

    public int getIndex(String fieldName) {
        boolean find = false;
        short i = -1;
        while ((!find) && (i < fields.length)) {
            i++;
            find = fields[i].getName().equals(fieldName);
        }
        if (!find) {
            throw new FdbSchemaException("no field with the name "+fieldName+" in class "+className);
        }
        return i;
    }

    public FdbFielComparator getComparator(int index) {
        return fields[index].getType().getComparator();
    }

    public InFileAddress getIndexIfa(int index) {
        return fields[index].getValueIndexIfa();
    }

    public void setIndexIfa(int index, InFileAddress ifa) {
        fields[index].setValueIndexIfa(ifa);
    }

    public short getIndexType(int index) {
        return fields[index].getValueIndexType();
    }

    public void setIndexType(int index, short indexType) {
        fields[index].setValueIndexType(indexType);
    }

    public Object getValueFor(String fieldName, Object[] values) {
        return values[getIndex(fieldName)];
    }

    /**
     * Return an array of the names of the declared fields of the described class.
     * @return String[] The names of the described class fields.
     */
    public String[] getFieldNames() {
        String[] result = new String[fields.length];
        for (int i = 0; i < result.length; i++) {
            result[i] = fields[i].getName();
        }
        return result;
    }

    /**
     * Return the number of field already declared for the described class.
     * @return int The number of declared fields.
     */
    public int getNbrFields() {
        return fields.length;
    }

    /**
     * Insert a new field in the description of the class.
     * @param fieldDescription The name of the field to insert.
     */
    public void addField(FieldDescription fieldDescription) {
        FieldDescription[] nfd = new FieldDescription[fields.length + 1];
        short index = fieldDescription.getIndex();
        if (index < 0) {
            index = (short) fields.length;
            fieldDescription.setIndex(index);
        }
        for (short i = 0; i < nfd.length; i++) {
            if (i < index) {
                nfd[i] = fields[i];
            } else if (i == index) {
                nfd[i] = fieldDescription;
            } else {
                fields[i - 1].setIndex(i);
                nfd[i] = fields[i - 1];
            }
        }
        fields = nfd;
    }

    /**
     * Remove a field from the description of the class.
     * @param index The index of the place where to remove the field.
     */
    public void removeField(int index) {
        FieldDescription[] nfd = new FieldDescription[fields.length - 1];
        for (short i = 0; i < nfd.length; i++) {
            if (i < index) {
                nfd[i] = fields[i];
            } else {
                fields[i + 1].setIndex(i);
                nfd[i] = fields[i + 1];
            }
        }
        fields = nfd;
    }

    public String toString() {
        return className + " with " + fields.length + " fields";
    }

    public String toStringFull(int nbrTab) {
        String tab = "";
        for (int i = 0; i < nbrTab; i++) {
            tab = tab + "\t";
        }
        StringBuffer sb = new StringBuffer();
        sb.append(tab).append(className).append("\n");
        for (int i = 0; i < fields.length; i++) {
            sb.append(tab).append("\t").append(fields[i]).append("\n");
        }
        return sb.toString();
    }

    public boolean equals(Object o) {
        try {
            ClassMetaDescription candidate = (ClassMetaDescription) o;
            boolean classTest = candidate.className.equals(this.className);
            boolean sizeTest = (candidate.fields.length == this.fields.length);
            if (classTest && sizeTest) {
                for (int i = 0; i < fields.length; i++) {
                    if (!candidate.fields[i].equals(this.fields[i])) {
                        return false;
                    }
                }
                return true;
            }
            return false;
        } catch (Exception e) {
        }
        return false;
    }

    private String className;
    private FieldDescription[] fields;
}
