package org.objectweb.fdb.jca;

import org.objectweb.fdb.impl.trace.Trace;
import org.objectweb.fdb.FdbSession;
import org.objectweb.fdb.FdbTransaction;
import org.objectweb.fdb.impl.meta.FdbMetaRepository;
import org.objectweb.fdb.impl.meta.ClassMetaDescription;
import org.objectweb.fdb.impl.util.FdbUtils;
import org.objectweb.fdb.api.FdbSessionInternal;
import org.objectweb.fdb.api.FdbQueryManager;
import org.objectweb.fdb.exception.FdbJcaException;

import javax.resource.cci.*;
import javax.resource.ResourceException;
import javax.resource.spi.ManagedConnection;
import javax.transaction.xa.XAException;
import java.util.Collection;

public class FdbConnection implements Connection, FdbSession {

    public FdbConnection(FdbManagedConnection fdbMc, Trace trace) {
        this.trace = trace;
        this.fdbMc = fdbMc;
    }

    public void setMc(ManagedConnection mc) {
        if (mc instanceof FdbManagedConnection) {
            this.fdbMc = (FdbManagedConnection) mc;
        } else {
            throw new FdbJcaException("Cannot manage connection of this class : " + mc.getClass().toString());
        }
    }

    public void open(FdbManagedConnection fdbMc) throws ResourceException {
        this.fdbMc = fdbMc;
        isValide = true;
        isClosed = false;
    }

    public void invalidate() throws ResourceException {
        fdbMc = null;
        isValide = false;
    }

    public boolean isClosed() {
        return isClosed;
    }



    private final FdbSession getFdbSession() {
        try {
            checkClosedOrNotValid();
        } catch (ResourceException e) {
            new FdbJcaException(CLOSED_EXCEPTION_MSG);
        }
        FdbSession result = fdbMc.getFdbSession();
        if (result == null) {
            throw new FdbJcaException("PM null in connection... You may be outside of a transaction");
        }
        return result;
    }

    private void checkClosedOrNotValid() throws ResourceException {
        if ((isClosed || (!isValide))) {
            throw new ResourceException(CLOSED_EXCEPTION_MSG);
        }
    }

    private String addressOf(Object o) {
        String s = String.valueOf(o);
        int index = s.indexOf('@');
        if (index == -1) return s;
        return FdbUtils.substring(s,index);
    }



    public String toString() {
        return "FdbConnection"+addressOf(super.toString())+" ManagedConnection"+addressOf(fdbMc)+" ManagedPM"+addressOf(fdbMc==null?null:fdbMc.getFdbSession());
    }

    /*************************** Connection implementation   **************************************/

    public LocalTransaction getLocalTransaction() throws ResourceException {
        checkClosedOrNotValid();
        return (javax.resource.cci.LocalTransaction) fdbMc.getLocalTransaction();
    }

    public synchronized void close() {
        trace.println(Trace.JCA, TRACE_ID + " closing connection");
        try {
            checkClosedOrNotValid();
        } catch (ResourceException e) {
            new FdbJcaException(CLOSED_EXCEPTION_MSG);
        }
        try {
            fdbMc.close(this);
        } catch (XAException e) {
            e.printStackTrace();
        }
        isClosed = true;
        trace.println(Trace.JCA, TRACE_ID + " connection closed...");
    }

    public Interaction createInteraction() throws ResourceException {
        checkClosedOrNotValid();
        return null;
    }

    public ConnectionMetaData getMetaData() throws ResourceException {
        checkClosedOrNotValid();
        return null;
    }

    public ResultSetInfo getResultSetInfo() throws ResourceException {
        checkClosedOrNotValid();
        return null;
    }

    public void setAutoCommit(boolean b) throws ResourceException {
        checkClosedOrNotValid();
    }

    public boolean getAutoCommit() throws ResourceException {
        checkClosedOrNotValid();
        return false;
    }

    /***************************   FdbSession Implementation ***************************************/

    public boolean contains(Object oid) {
        return getFdbSession().contains(oid);
    }

    public Object createObject(Object[] objectToCreate, Class objectClass) {
        return getFdbSession().createObject(objectToCreate, objectClass);
    }

    public Object createObject(Object[] objectToCreate, String objectClassName) {
        return getFdbSession().createObject(objectToCreate, objectClassName);
    }

    public Object createObject(Object oid, Object[] objectToCreate) {
        return getFdbSession().createObject(oid, objectToCreate);
    }

    public FdbTransaction currentTransaction() {
        return getFdbSession().currentTransaction();
    }

    public void defineClass(ClassMetaDescription classDescription) {
        getFdbSession().defineClass(classDescription);
    }

    public void deleteObjectByOid(Object oid) {
        getFdbSession().deleteObjectByOid(oid);
    }

    public String getClassNameFor(Object oid) {
        return null;
    }

    public Collection getExtent(String fullCLassName) {
        return getFdbSession().getExtent(fullCLassName);
    }

    public Collection getExtent(Class theClass) {
        return getFdbSession().getExtent(theClass);
    }

    public FdbSessionInternal getInternalSession() {
        return getFdbSession().getInternalSession();
    }

    public FdbQueryManager getQueryManager() {
        return getFdbSession().getQueryManager();
    }

    public FdbMetaRepository getMetaRepository() {
        return getFdbSession().getMetaRepository();
    }

    public boolean isClassDefined(String fullQualifiedClassName) {
        return getFdbSession().isClassDefined(fullQualifiedClassName);
    }

    public boolean isNewBase() {
        return getFdbSession().isNewBase();
    }

    public Object makeNewFileOid(Class objectClass) {
        return getFdbSession().makeNewFileOid(objectClass);
    }

    public Object makeNewFileOid(String objectClassName) {
        return getFdbSession().makeNewFileOid(objectClassName);
    }

    public Object[] readObjectByOid(Object oid, boolean withCache) {
        return getFdbSession().readObjectByOid(oid, withCache);
    }

    public void restructureBase() {
        getFdbSession().restructureBase();
    }

    public Object updateObjectByOid(Object oid, Object[] objectToUpdate) {
        return getFdbSession().updateObjectByOid(oid, objectToUpdate);
    }


    private static final String TRACE_ID = "[FdbConnection]";
    private static final String CLOSED_EXCEPTION_MSG = "Connection already closed or invalidated...";
    private FdbManagedConnection fdbMc;
    private Trace trace;
    private boolean isClosed = false;
    private boolean isValide = true;
}
