package org.objectweb.fdb.jca;

import org.objectweb.fdb.FdbSession;
import org.objectweb.fdb.exception.FdbJcaException;
import org.objectweb.fdb.impl.trace.Trace;
import org.objectweb.fdb.impl.trace.TraceNull;
import org.objectweb.fdb.jca.data.FdbManagedConnectionMetaData;

import javax.resource.spi.*;
import javax.resource.ResourceException;
import javax.security.auth.Subject;
import javax.transaction.xa.XAResource;
import javax.transaction.xa.XAException;
import java.util.List;
import java.util.ArrayList;
import java.io.PrintWriter;

public class FdbManagedConnection implements ManagedConnection {

    public FdbManagedConnection(String connectionURL, ConnectionRequestInfo info, Trace trace) {
        this.connectionURL = connectionURL;
        this.trace = trace;
        this.info = info;
        metadata = new FdbManagedConnectionMetaData();
        try {
            getXAResource();
        } catch (ResourceException e) {
            throw new FdbJcaException("Can't initialize the FdbManagedConnection",e);
        }
    }

    public void setManagedFdbSession(ManagedFdbSession mFdbSession) {
        this.mFdbSession = mFdbSession;
    }

    public boolean isAvalaible() {
        return (isAvalaible && (mFdbSession == null) && jdoxarWrapper.isAvailable());
    }

    public void setNotAvalaible() {
        isAvalaible = false;
    }

    public void setCfIdentification(String id) {
        cfIdentification = id;
    }

    public String getCfIdentification() {
        return cfIdentification;
    }

    public String getConnectionURL() {
        return connectionURL;
    }


    public void close(FdbConnection c) throws XAException {
        trace("close(c)");
        if (!c.equals(activeConnection)) {
            throw new XAException("Cannot close a not active connection...");
        }
        for (int i = 0; i < eventListeners.size(); i++) {
            ConnectionEventListener cel = (ConnectionEventListener) eventListeners.get(i);
            ConnectionEvent ce = new ConnectionEvent(this, ConnectionEvent.CONNECTION_CLOSED);
            ce.setConnectionHandle(c);
            cel.connectionClosed(ce);
        }
    }

    public FdbSession getFdbSession() {
        return mFdbSession;
    }



//    // propagate an error on the instance's active connection
//    public void error(Exception e) throws XAException {
//        trace("error(e) : " + e.getMessage());
//        for (int i = 0; i < eventListeners.size(); i++) {
//            ConnectionEventListener cel = (ConnectionEventListener) eventListeners.get(i);
//            ConnectionEvent ce = new ConnectionEvent(this, ConnectionEvent.CONNECTION_ERROR_OCCURRED);
//            ce.setConnectionHandle(activeConnection);
//            cel.connectionErrorOccurred(ce);
//        }
//    }



    public boolean matchConnectionRequestInfo(ConnectionRequestInfo requestInfo) {
        if (info == null) {
            return (requestInfo == null);
        }
        // info != null
        return info.equals(requestInfo);
    }

    private void trace(Object o) {
        String message = TRACE_ID + " " + String.valueOf(o);
        trace.println(Trace.JCA, message);
    }



    public boolean equals(Object o) {
        boolean result = false;
        if (o instanceof FdbManagedConnection) {
            FdbManagedConnection candidat = (FdbManagedConnection) o;
            result = (this.hashCode() == candidat.hashCode());
        }
        return result;
    }

    public int hashCode() {
        if (hashCode == 0) {
            hashCode = super.hashCode();
        }
        return hashCode;
    }


    /*****************************    ManagedConnection implementation ****************************/

    // pas de connection sharing pour l'instant (p40 et 71 de la spec)
    public Object getConnection(Subject subject, ConnectionRequestInfo info) throws ResourceException {
        trace("getConnection");
        if (activeConnection == null) {
            activeConnection = new FdbConnection(this, trace);
        } else {
            if (activeConnection.isClosed()) {
                activeConnection.open(this);
            }
            //throw new JDOUserException("not the first call to getConnection, but don't support connection sharing");
        }
        return activeConnection;
    }

    public void associateConnection(Object c) throws ResourceException {
        trace("associateConnection " + c.getClass().toString());

        if (c == null) {
            throw new ResourceException("Cannot associate a null connection");
        }

        if (activeConnection == null) {
            activeConnection = (FdbConnection) c;
            return;
        }

        if (activeConnection.equals(c)) {
            return;
        }

        if (c instanceof FdbConnection) {
            FdbConnection jdoConnection = (FdbConnection) c;
            jdoConnection.setMc(this);
            activeConnection.invalidate();
            activeConnection = jdoConnection;
        } else {
            throw new ResourceException("The managedConnection cannot manage this connection : not a JDOConnection");
        }
    }

    public XAResource getXAResource() throws ResourceException {
        trace("getXAResource");
        if (jdoxarWrapper == null) {
            trace("getXAResource : new ressource");
            jdoxarWrapper = new FdbXAResourceWrapper(this, trace);
        }
        return jdoxarWrapper;
    }

    public void addConnectionEventListener(ConnectionEventListener listener) {
        trace("addConnectionEventListener");
        eventListeners.add(listener);
    }

    public void removeConnectionEventListener(ConnectionEventListener listener) {
        trace("removeConnectionEventListener");
        eventListeners.remove(listener);
    }

    public void cleanup() throws ResourceException {
        trace("cleanup");
        if (activeConnection != null) {
            activeConnection.invalidate();
            activeConnection = null;
        }
        jdoxarWrapper.cleanUp();
        mFdbSession = null;
        isAvalaible = true;
    }

    public void destroy() throws ResourceException {
        trace("destroy");
        jdoxarWrapper = null;
        isAvalaible = false;
        eventListeners.clear();
        logger = null;
    }

    // must be spi.LocalTransaction
    public LocalTransaction getLocalTransaction() throws ResourceException {
        throw new RuntimeException("not implemented yet...");
        //		return mFdbSession.getLocalManagedTransaction();
    }

    public ManagedConnectionMetaData getMetaData() throws ResourceException {
        trace("getMetaData : " + metadata.toString());
        return metadata;
    }

    public void setLogWriter(PrintWriter writer) throws ResourceException {
        trace("setLogWriter");
        logger = writer;
    }

    public PrintWriter getLogWriter() throws ResourceException {
        trace("getLogWriter");
        return logger;
    }


    private ManagedFdbSession mFdbSession;
    private FdbConnection activeConnection = null;
    private FdbXAResourceWrapper jdoxarWrapper;

    private String connectionURL;
    private String cfIdentification;
    private ConnectionRequestInfo info;
    private FdbManagedConnectionMetaData metadata;

    private List eventListeners = new ArrayList();

    private PrintWriter logger;
    private Trace trace = new TraceNull();


    private boolean isAvalaible = true;
    private int hashCode = 0;

    private static final String TRACE_ID = "[FdbManagedConnection]";
}
