package fdbtests.workbench;

import junit.extensions.jfunc.AssertListener;
import junit.extensions.jfunc.textui.JFuncRunner;
import junit.extensions.jfunc.util.ColorWriter;

import junit.framework.AssertionFailedError;
import junit.framework.Test;
import junit.framework.TestResult;

import java.io.IOException;
import java.io.InputStream;
import java.io.File;

import java.util.*;

import fdbtests.workbench.FdbTestCase;

public class FdbTestEngine extends JFuncRunner {
    static final String NO_PROPS_FILE = "The properties file for the database can't be located in classpath";
    private Properties dbProperties;
    private String dbPropertiesFilename;
    private boolean isSuiteAlreadyInit = false;

    public FdbTestEngine(String dbProp) {
        this.dbPropertiesFilename = dbProp;
    }

    public Properties getDBProperties() {
        return dbProperties;
    }

    public String getDBPropertiesFilename() {
        return dbPropertiesFilename;
    }

    public TestResult doRun(Test suite, boolean wait) {
        TestResult result = null;

        try {
            ((FdbTestSuite) suite).setTestEngine(this);

            TestResult result11 = super.createTestResult();
            result11.addListener(new VerboseListener());

            long startTime = System.currentTimeMillis();
            suite.run(result11);

            long endTime = System.currentTimeMillis();
            long runTime = endTime - startTime;
            writer().println();

            writer().println("Time: " + super.elapsedTimeAsString(runTime));
            print(result11);

            writer().println();

            pause(wait);
            result = result11;
        } catch (Exception e) {
            e.printStackTrace();
        }

        return result;
    }

    public static void main(String[] args) {
        String dbProp = null;
        String nameTestSuite = null;

        if (args.length == 0) {
            printHelp();
        }

        List newArgList = new ArrayList();

        for (int i = 0; i < args.length; i++) {
            if (args[i].equals("--help") || args[i].equals("-h")) {
                printHelp();
            } else if (args[i].equals("-db")) {
                i = i + 1;
                dbProp = args[i];
            } else if (args[i].equals("-nt")) {
                i = i + 1;
                nameTestSuite = args[i];
            } else {
                newArgList.add(args[i]);
            }
        }

        FdbTestEngine aTestRunner = new FdbTestEngine(dbProp);

        newArgList.add("-c");
        newArgList.add(nameTestSuite);

        String[] newArgs = (String[]) newArgList.toArray(new String[newArgList.size()]);

        try {
            TestResult r = aTestRunner.start(newArgs);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public void printErrors(TestResult testResult) {
        super.printErrors(testResult);
    }

    protected void suiteInit(FdbTestSuite testSuite) {
        if (!isSuiteAlreadyInit) {
            //Retrieve the database properties file
            ClassLoader classLoader = Thread.currentThread().getContextClassLoader();
            InputStream testIS = classLoader.getResourceAsStream(dbPropertiesFilename);
            Properties props = new Properties();

            try {
                props.load(testIS);
            } catch (IOException e) {
                System.out.println(NO_PROPS_FILE);
                e.printStackTrace();
            }

            if (testIS == null) {
                System.out.println(NO_PROPS_FILE);
                props = null;
            }

            dbProperties = props;
            isSuiteAlreadyInit = true;
        }
    }

    protected void testCaseInit(FdbTestCase testCase) {
    }

    private static void printHelp() {
        System.out.println("Parameters :");
        System.out.println(" -db <dbFile> : the name of the database properties file");
        System.out.println(" -nt <nameTest> : the name of the test class");
        System.out.println(" -h or --help : diplay this help");
        System.out.println("Options of jfunc : \n" +
                           "[--wait] [-v] [--color] testCaseName [args]\n" +
                           "  --version   display version\n" +
                           "  --wait      wait between tests\n" +
                           "  -v          be verbose (show verbose assertions)\n" +
                           "  --color     display ANSI colors");
        System.out.println("All files must be locate in classpath.");
        System.exit(0);
    }

    class VerboseListener implements AssertListener {
        ColorWriter out = new ColorWriter(writer());

        public VerboseListener() {
            out.enableColor(useColor());
        }

        public void addAssert(Test test, String msg, boolean condition) {
            int color;
            String passorfail;

            if (!condition) {
                // if failed
                return;
            }

            if (condition) {
                color = ColorWriter.GREEN;
                passorfail = "PASSED";
            } else {
                color = ColorWriter.RED;
                passorfail = "FAILED";
            }

            status(color, passorfail, msg);
        }

        public void addError(Test test, Throwable t) {
            //status(ColorWriter.BLACK, ColorWriter.YELLOW, "ERROR");
            status(ColorWriter.YELLOW, "ERROR", t.toString());

            //out.println("error: " + test + " " + t);
        }

        public void addFailure(Test test, AssertionFailedError t) {
            status(ColorWriter.RED, "FAILED", t.getMessage());

            //out.println("failure: " + test + " " + t);
        }

        public void endTest(Test test) {
            //out.println("end: " + test);
        }

        public void startTest(Test test) {
            out.setColor(ColorWriter.CYAN);
            out.println(test.toString() + ": ");
            out.setColor(ColorWriter.DEFAULT);
        }

        private void status(int color, String coloredMsg, String msg) {
            out.print("  [ ");
            out.setColor(color);
            out.print(coloredMsg);
            out.setColor(ColorWriter.DEFAULT);
            out.print(" " + msg);
            out.println(" ]");
        }
    }
}
